# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

# TODO
# 1: reorganize into private and public methods
# 2: Include smoothing?

import os
import pickle

import numpy as np
import pandas as pd
import dateutil
import matplotlib.cm as cm
import matplotlib.pyplot as plt
# from pyproj import Proj

import floris.utilities as geo


class WindRose:
    """
    The WindRose class is used to organize information about the frequency of
    occurance of different combinations of wind speed and wind direction (and
    other optimal wind variables). A WindRose object can be used to help
    calculate annual energy production (AEP) when combined with Floris power
    calculations for different wind conditions. Several methods exist for
    populating a WindRose object with wind data. WindRose also contains methods
    for visualizing wind roses.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: wr-
    """

    def __init__(self,):
        """
        Instantiate a WindRose object and set some initial parameter values.
        No explicit arguments required, and an additional method will need to
        be called to populate the WindRose object with data.
        """
        # Initialize some varibles
        self.num_wd = 0
        self.num_ws = 0
        self.wd_step = 1.0
        self.ws_step = 5.0
        self.wd = np.array([])
        self.ws = np.array([])
        self.df = pd.DataFrame()

    def save(self, filename):
        """
        This method saves the WindRose data as a pickle file so that it can be
        imported into a WindRose object later.

        Args:
            filename (str): Path and filename of pickle file to save.
        """
        pickle.dump(
            [
                self.num_wd,
                self.num_ws,
                self.wd_step,
                self.ws_step,
                self.wd,
                self.ws,
                self.df,
            ],
            open(filename, "wb"),
        )

    def load(self, filename):
        """
        This method loads data from a previously saved WindRose pickle file
        into a WindRose object.

        Args:
            filename (str): Path and filename of pickle file to load.

        Returns:
            int, int, float, float, np.array, np.array, pandas.DataFrame:

                -   Number of wind direction bins.
                -   Number of wind speed bins.
                -   Wind direction bin size (deg).
                -   Wind speed bin size (m/s).
                -   List of wind direction bin center values (deg).
                -   List of wind speed bin center values (m/s).
                -   DataFrame containing at least the following columns:

                    - **wd** (*float*) - Wind direction bin center values (deg).
                    - **ws** (*float*) - Wind speed bin center values (m/s).
                    - **freq_val** (*float*) - The frequency of occurance of
                      the wind conditions in the other columns.
        """
        (
            self.num_wd,
            self.num_ws,
            self.wd_step,
            self.ws_step,
            self.wd,
            self.ws,
            self.df,
        ) = pickle.load(open(filename, "rb"))

        return self.df

    def resample_wind_speed(self, df, ws=np.arange(0, 26, 1.0)):
        """
        This method resamples the wind speed bins using the specified wind
        speed bin center values. The frequency values are adjusted accordingly.

        Args:
            df (pandas.DataFrame): Wind rose DataFrame containing at least the
                following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  wind conditions in the other columns.

            ws (np.array, optional): List of new wind speed center bins (m/s).
                Defaults to np.arange(0, 26, 1.).

        Returns:
            pandas.DataFrame: Wind rose DataFrame with the resampled wind speed
            bins and frequencies containing at least the following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - New wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  new wind conditions in the other columns.
        """
        # Make a copy of incoming dataframe
        df = df.copy(deep=True)

        # Get the wind step
        ws_step = ws[1] - ws[0]

        # Ws
        ws_edges = ws - ws_step / 2.0
        ws_edges = np.append(ws_edges, np.array(ws[-1] + ws_step / 2.0))

        # Cut wind speed onto bins
        df["ws"] = pd.cut(df.ws, ws_edges, labels=ws)

        # Regroup
        df = df.groupby([c for c in df.columns if c != "freq_val"]).sum()

        # Fill nans
        df = df.fillna(0)

        # Reset the index
        df = df.reset_index()

        # Set to float
        for c in [c for c in df.columns if c != "freq_val"]:
            df[c] = df[c].astype(float)
            df[c] = df[c].astype(float)

        return df

    def internal_resample_wind_speed(self, ws=np.arange(0, 26, 1.0)):
        """
        Internal method for resampling wind speed into desired bins. The
        frequency values are adjusted accordingly. Modifies data within
        WindRose object without explicit return.

        TODO: make a private method

        Args:
            ws (np.array, optional): Vector of wind speed bin centers for
                the wind rose (m/s). Defaults to np.arange(0, 26, 1.).
        """
        # Update ws and wd binning
        self.ws = ws
        self.num_ws = len(ws)
        self.ws_step = ws[1] - ws[0]

        # Update internal data frame
        self.df = self.resample_wind_speed(self.df, ws)

    def resample_wind_direction(self, df, wd=np.arange(0, 360, 5.0)):
        """
        This method resamples the wind direction bins using the specified wind
        direction bin center values. The frequency values are adjusted
        accordingly.

        Args:
            df (pandas.DataFrame): Wind rose DataFrame containing at least the
                following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  wind conditions in the other columns.

            wd (np.array, optional): List of new wind direction center bins
                (deg). Defaults to np.arange(0, 360, 5.).

        Returns:
            pandas.DataFrame: Wind rose DataFrame with the resampled wind
            direction bins and frequencies containing at least the following
            columns:

                - **wd** (*float*) - New wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  new wind conditions in the other columns.
        """
        # Make a copy of incoming dataframe
        df = df.copy(deep=True)

        # Get the wind step
        wd_step = wd[1] - wd[0]

        # Get bin edges
        wd_edges = wd - wd_step / 2.0
        wd_edges = np.append(wd_edges, np.array(wd[-1] + wd_step / 2.0))

        # Get the overhangs
        negative_overhang = wd_edges[0]
        positive_overhang = wd_edges[-1] - 360.0

        # Need potentially to wrap high angle direction to negative for correct
        # binning
        df["wd"] = geo.wrap_360(df.wd)
        if negative_overhang < 0:
            print("Correcting negative Overhang:%.1f" % negative_overhang)
            df["wd"] = np.where(
                df.wd.values >= 360.0 + negative_overhang,
                df.wd.values - 360.0,
                df.wd.values,
            )

        # Check on other side
        if positive_overhang > 0:
            print("Correcting positive Overhang:%.1f" % positive_overhang)
            df["wd"] = np.where(
                df.wd.values <= positive_overhang, df.wd.values + 360.0, df.wd.values
            )

        # Cut into bins
        df["wd"] = pd.cut(df.wd, wd_edges, labels=wd)

        # Regroup
        df = df.groupby([c for c in df.columns if c != "freq_val"]).sum()

        # Fill nans
        df = df.fillna(0)

        # Reset the index
        df = df.reset_index()

        # Set to float Re-wrap
        for c in [c for c in df.columns if c != "freq_val"]:
            df[c] = df[c].astype(float)
            df[c] = df[c].astype(float)
        df["wd"] = geo.wrap_360(df.wd)

        return df

    def internal_resample_wind_direction(self, wd=np.arange(0, 360, 5.0)):
        """
        Internal method for resampling wind direction into desired bins. The
        frequency values are adjusted accordingly. Modifies data within
        WindRose object without explicit return.

        TODO: make a private method

        Args:
            wd (np.array, optional): Vector of wind direction bin centers for
                the wind rose (deg). Defaults to np.arange(0, 360, 5.).
        """
        # Update ws and wd binning
        self.wd = wd
        self.num_wd = len(wd)
        self.wd_step = wd[1] - wd[0]

        # Update internal data frame
        self.df = self.resample_wind_direction(self.df, wd)

    def resample_column(self, df, col, bins):
        """
        This method resamples the specified wind parameter column using the
        specified bin center values. The frequency values are adjusted
        accordingly.

        Args:
            df (pandas.DataFrame): Wind rose DataFrame containing at least the
                following columns as well as *col*:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  wind conditions in the other columns.

            col (str): The name of the column to resample.
            bins (np.array): List of new bin center values for the specified
                column.

        Returns:
            pandas.DataFrame: Wind rose DataFrame with the resampled wind
            parameter bins and frequencies containing at least the following
            columns as well as *col*:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  new wind conditions in the other columns.
        """
        # Make a copy of incoming dataframe
        df = df.copy(deep=True)

        # Cut into bins, make first and last bins extend to -/+ infinity
        var_edges = np.append(0.5 * (bins[1:] + bins[:-1]), np.inf)
        var_edges = np.append(-np.inf, var_edges)
        df[col] = pd.cut(df[col], var_edges, labels=bins)

        # Regroup
        df = df.groupby([c for c in df.columns if c != "freq_val"]).sum()

        # Fill nans
        df = df.fillna(0)

        # Reset the index
        df = df.reset_index()

        # Set to float
        for c in [c for c in df.columns if c != "freq_val"]:
            df[c] = df[c].astype(float)

        return df

    def internal_resample_column(self, col, bins):
        """
        Internal method for resampling column into desired bins. The frequency
        values are adjusted accordingly. Modifies data within WindRose object
        without explicit return.

        TODO: make a private method

        Args:
            col (str): Name of column to resample.
            bins (np.array): Vector of bins for the WindRose column.
        """
        # Update internal data frame
        self.df = self.resample_column(self.df, col, bins)

    def resample_average_ws_by_wd(self, df):
        """
        This method calculates the mean wind speed for each wind direction bin
        and resamples the wind rose, resulting in a single mean wind speed per
        wind direction bin. The frequency values are adjusted accordingly.

        Args:
            df (pandas.DataFrame): Wind rose DataFrame containing at least the
                following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  wind conditions in the other columns.

        Returns:
            pandas.DataFrame: Wind rose DataFrame with the resampled wind speed
            bins and frequencies containing at least the following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - The average wind speed for each wind
                  direction bin (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  new wind conditions in the other columns.
        """
        # Make a copy of incoming dataframe
        df = df.copy(deep=True)

        ws_avg = []

        for val in df.wd.unique():
            ws_avg.append(
                np.array(
                    df.loc[df["wd"] == val]["ws"] * df.loc[df["wd"] == val]["freq_val"]
                ).sum()
                / df.loc[df["wd"] == val]["freq_val"].sum()
            )

        # Regroup
        df = df.groupby("wd").sum()

        df["ws"] = ws_avg

        # Reset the index
        df = df.reset_index()

        # Set to float
        df["ws"] = df.ws.astype(float)
        df["wd"] = df.wd.astype(float)

        return df

    def internal_resample_average_ws_by_wd(self, wd=np.arange(0, 360, 5.0)):
        """
        This internal method calculates the mean wind speed for each specified
        wind direction bin and resamples the wind rose, resulting in a single
        mean wind speed per wind direction bin. The frequency values are
        adjusted accordingly.

        TODO: make an internal method

        Args:
            wd (np.arange, optional): Wind direction bin centers (deg).
            Defaults to np.arange(0, 360, 5.).

        Returns:
            pandas.DataFrame: Wind rose DataFrame with the resampled wind speed
            bins and frequencies containing at least the following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - The average wind speed for each wind
                  direction bin (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  new wind conditions in the other columns.
        """
        # Update ws and wd binning
        self.wd = wd
        self.num_wd = len(wd)
        self.wd_step = wd[1] - wd[0]

        # Update internal data frame
        self.df = self.resample_average_ws_by_wd(self.df)

    def weibull(self, x, k=2.5, lam=8.0):
        """
        This method returns a Weibull distribution corresponding to the input
        data array (typically wind speed) using the specified Weibull
        parameters.

        Args:
            x (np.array): List of input data (typically binned wind speed
                observations).
            k (float, optional): Weibull shape parameter. Defaults to 2.5.
            lam (float, optional): Weibull scale parameter. Defaults to 8.0.

        Returns:
            np.array: Weibull distribution probabilities corresponding to
            values in the input array.
        """
        return (k / lam) * (x / lam) ** (k - 1) * np.exp(-((x / lam) ** k))

    def make_wind_rose_from_weibull(
        self, wd=np.arange(0, 360, 5.0), ws=np.arange(0, 26, 1.0)
    ):
        """
        Populate WindRose object with an example wind rose with wind speed
        frequencies given by a Weibull distribution. The wind direction
        frequencies are initialized according to an example distribution.

        Args:
            wd (np.array, optional): Wind direciton bin centers (deg). Defaults
            to np.arange(0, 360, 5.).
            ws (np.array, optional): Wind speed bin centers (m/s). Defaults to
                np.arange(0, 26, 1.).

        Returns:
            pandas.DataFrame: Wind rose DataFrame containing at least the
            following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  wind conditions in the other columns.
        """
        # Use an assumed wind-direction for dir frequency
        wind_dir = [
            0,
            22.5,
            45,
            67.5,
            90,
            112.5,
            135,
            157.5,
            180,
            202.5,
            225,
            247.5,
            270,
            292.5,
            315,
            337.5,
        ]
        freq_dir = [
            0.064,
            0.04,
            0.038,
            0.036,
            0.045,
            0.05,
            0.07,
            0.08,
            0.11,
            0.08,
            0.05,
            0.036,
            0.048,
            0.058,
            0.095,
            0.10,
        ]

        freq_wd = np.interp(wd, wind_dir, freq_dir)
        freq_ws = self.weibull(ws)

        freq_tot = np.zeros(len(wd) * len(ws))
        wd_tot = np.zeros(len(wd) * len(ws))
        ws_tot = np.zeros(len(wd) * len(ws))

        count = 0
        for i in range(len(wd)):
            for j in range(len(ws)):
                wd_tot[count] = wd[i]
                ws_tot[count] = ws[j]

                freq_tot[count] = freq_wd[i] * freq_ws[j]
                count = count + 1

        # renormalize
        freq_tot = freq_tot / np.sum(freq_tot)

        # Load the wind toolkit data into a dataframe
        df = pd.DataFrame()

        # Start by simply round and wrapping the wind direction and wind speed
        # columns
        df["wd"] = wd_tot
        df["ws"] = ws_tot

        # Now group up
        df["freq_val"] = freq_tot

        # Save the df at this point
        self.df = df
        # TODO is there a reason self.df is updated AND returned?
        return self.df

    def make_wind_rose_from_user_data(
        self, wd_raw, ws_raw, *args, wd=np.arange(0, 360, 5.0), ws=np.arange(0, 26, 1.0)
    ):
        """
        This method populates the WindRose object given user-specified
        observations of wind direction, wind speed, and additional optional
        variables. The wind parameters are binned and the frequencies of
        occurance of each binned wind condition combination are calculated.

        Args:
            wd_raw (array-like): An array-like list of all wind direction
                observations used to calculate the normalized frequencies (deg).
            ws_raw (array-like): An array-like list of all wind speed
                observations used to calculate the normalized frequencies (m/s).
            *args: Variable length argument list consisting of a sequence of
                the following alternating arguments:

                -   string - Name of additional wind parameters to include in
                    wind rose.
                -   array-like - Values of the additional wind parameters used
                    to calculate the frequencies of occurance
                -   np.array - Bin center values for binning the additional
                    wind parameters.

            wd (np.array, optional): Wind direction bin centers (deg). Defaults
                to np.arange(0, 360, 5.).
            ws (np.array, optional): Wind speed bin limits (m/s). Defaults to
                np.arange(0, 26, 1.).

        Returns:
            pandas.DataFrame: Wind rose DataFrame containing at least the
            following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  wind conditions in the other columns.
        """
        df = pd.DataFrame()

        # convert inputs to np.array
        wd_raw = np.array(wd_raw)
        ws_raw = np.array(ws_raw)

        # Start by simply round and wrapping the wind direction and wind speed
        # columns
        df["wd"] = geo.wrap_360(wd_raw.round())
        df["ws"] = ws_raw.round()

        # Loop through *args and assign new dataframe columns after cutting
        # into possibly irregularly-spaced bins
        for in_var in range(0, len(args), 3):
            df[args[in_var]] = np.array(args[in_var + 1])

            # Cut into bins, make first and last bins extend to -/+ infinity
            var_edges = np.append(
                0.5 * (args[in_var + 2][1:] + args[in_var + 2][:-1]), np.inf
            )
            var_edges = np.append(-np.inf, var_edges)
            df[args[in_var]] = pd.cut(
                df[args[in_var]], var_edges, labels=args[in_var + 2]
            )

        # Now group up
        df["freq_val"] = 1.0
        df = df.groupby([c for c in df.columns if c != "freq_val"]).sum()
        df["freq_val"] = df.freq_val.astype(float) / df.freq_val.sum()
        df = df.reset_index()

        # Save the df at this point
        self.df = df

        # Resample onto the provided wind speed and wind direction binnings
        self.internal_resample_wind_speed(ws=ws)
        self.internal_resample_wind_direction(wd=wd)

        return self.df

    def make_wind_rose_from_user_dist(
        self,
        wd_raw,
        ws_raw,
        freq_val,
        *args,
        wd=np.arange(0, 360, 5.0),
        ws=np.arange(0, 26, 1.0),
    ):
        """
        This method populates the WindRose object given user-specified
        combinations of wind direction, wind speed, additional optional
        variables, and the corresponding frequencies of occurance. The wind
        parameters are binned using the specified wind parameter bin center
        values and the corresponding frequencies of occrance are calculated.

        Args:
            wd_raw (array-like): An array-like list of wind directions
                corresponding to the specified frequencies of occurance (deg).
            wd_raw (array-like): An array-like list of wind speeds
                corresponding to the specified frequencies of occurance (m/s).
            freq_val (array-like): An array-like list of normalized frequencies
                corresponding to the provided wind parameter combinations.
            *args: Variable length argument list consisting of a sequence of
                the following alternating arguments:

                -   string - Name of additional wind parameters to include in
                    wind rose.
                -   array-like - Values of the additional wind parameters
                    corresponding to the specified frequencies of occurance.
                -   np.array - Bin center values for binning the additional
                    wind parameters.

            wd (np.array, optional): Wind direction bin centers (deg). Defaults
                to np.arange(0, 360, 5.).
            ws (np.array, optional): Wind speed bin centers (m/s). Defaults to
                np.arange(0, 26, 1.).

        Returns:
            pandas.DataFrame: Wind rose DataFrame containing at least the
            following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  wind conditions in the other columns.
        """
        df = pd.DataFrame()

        # convert inputs to np.array
        wd_raw = np.array(wd_raw)
        ws_raw = np.array(ws_raw)

        # Start by simply wrapping the wind direction column
        df["wd"] = geo.wrap_360(wd_raw)
        df["ws"] = ws_raw

        # Loop through *args and assign new dataframe columns
        for in_var in range(0, len(args), 3):
            df[args[in_var]] = np.array(args[in_var + 1])

        # Assign frequency column
        df["freq_val"] = np.array(freq_val)
        df["freq_val"] = df["freq_val"] / df["freq_val"].sum()

        # Save the df at this point
        self.df = df

        # Resample onto the provided wind variable binnings
        self.internal_resample_wind_speed(ws=ws)
        self.internal_resample_wind_direction(wd=wd)

        # Loop through *args and resample using provided binnings
        for in_var in range(0, len(args), 3):
            self.internal_resample_column(args[in_var], args[in_var + 2])

        return self.df

    def parse_wind_toolkit_folder(
        self,
        folder_name,
        wd=np.arange(0, 360, 5.0),
        ws=np.arange(0, 26, 1.0),
        limit_month=None,
    ):
        """
        This method populates the WindRose object given raw wind direction and
        wind speed data saved in csv files downloaded from the WIND Toolkit
        application (see https://www.nrel.gov/grid/wind-toolkit.html for more
        information). The wind parameters are binned using the specified wind
        parameter bin center values and the corresponding frequencies of
        occurance are calculated.

        Args:
            folder_name (str): Path to the folder containing the WIND Toolkit
                data files.
            wd (np.array, optional): Wind direction bin centers (deg). Defaults
                to np.arange(0, 360, 5.).
            ws (np.array, optional): Wind speed bin centers (m/s). Defaults to
                np.arange(0, 26, 1.).
            limit_month (list, optional): List of ints of month(s) (e.g., 1, 2
                3...) to consider when calculating the wind condition
                frequencies. If none are specified, all months will be used.
                Defaults to None.

        Returns:
            pandas.DataFrame: Wind rose DataFrame containing the following
            columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  wind conditions in the other columns.
        """
        # Load the wind toolkit data into a dataframe
        df = self.load_wind_toolkit_folder(folder_name, limit_month=limit_month)

        # Start by simply round and wrapping the wind direction and wind speed
        # columns
        df["wd"] = geo.wrap_360(df.wd.round())
        df["ws"] = geo.wrap_360(df.ws.round())

        # Now group up
        df["freq_val"] = 1.0
        df = df.groupby(["ws", "wd"]).sum()
        df["freq_val"] = df.freq_val.astype(float) / df.freq_val.sum()
        df = df.reset_index()

        # Save the df at this point
        self.df = df

        # Resample onto the provided wind speed and wind direction binnings
        self.internal_resample_wind_speed(ws=ws)
        self.internal_resample_wind_direction(wd=wd)

        return self.df

    def load_wind_toolkit_folder(self, folder_name, limit_month=None):
        """
        This method imports raw wind direction and wind speed data saved in csv
        files in the specified folder downloaded from the WIND Toolkit
        application (see https://www.nrel.gov/grid/wind-toolkit.html for more
        information).

        TODO: make private method?

        Args:
            folder_name (str): Path to the folder containing the WIND Toolkit
                csv data files.
            limit_month (list, optional): List of ints of month(s) (e.g., 1, 2,
                3...) to consider when calculating the wind condition
                frequencies. If none are specified, all months will be used.
                Defaults to None.

        Returns:
            pandas.DataFrame: DataFrame containing the following columns:

                - **wd** (*float*) - Raw wind direction data (deg).
                - **ws** (*float*) - Raw wind speed data (m/s).
        """
        file_list = os.listdir(folder_name)
        file_list = [os.path.join(folder_name, f) for f in file_list if ".csv" in f]

        df = pd.DataFrame()
        for f_idx, f in enumerate(file_list):
            print("%d of %d: %s" % (f_idx, len(file_list), f))
            df_temp = self.load_wind_toolkit_file(f, limit_month=limit_month)
            df = df.append(df_temp)

        return df

    def load_wind_toolkit_file(self, filename, limit_month=None):
        """
        This method imports raw wind direction and wind speed data saved in the
        specified csv file downloaded from the WIND Toolkit application (see
        https://www.nrel.gov/grid/wind-toolkit.html for more information).

        TODO: make private method?

        Args:
            filename (str): Path to the WIND Toolkit csv file.
            limit_month (list, optional): List of ints of month(s) (e.g., 1, 2,
                3...) to consider when calculating the wind condition
                frequencies. If none are specified, all months will be used.
                Defaults to None.

        Returns:
            pandas.DataFrame: DataFrame containing the following columns with
            data from the WIND Toolkit file:

                - **wd** (*float*) - Raw wind direction data (deg).
                - **ws** (*float*) - Raw wind speed data (m/s).
        """
        df = pd.read_csv(filename, header=3, sep=",")

        # If asked to limit to particular months
        if limit_month is not None:
            df = df[df.Month.isin(limit_month)]

        # Save just what I want
        speed_column = [c for c in df.columns if "speed" in c][0]
        direction_column = [c for c in df.columns if "direction" in c][0]
        df = df.rename(index=str, columns={speed_column: "ws", direction_column: "wd"})[
            ["wd", "ws"]
        ]

        return df

    def import_from_wind_toolkit_hsds(
        self,
        lat,
        lon,
        ht=100,
        wd=np.arange(0, 360, 5.0),
        ws=np.arange(0, 26, 1.0),
        include_ti=False,
        limit_month=None,
        limit_hour=None,
        st_date=None,
        en_date=None,
    ):
        """
        This method populates the WindRose object using wind data from the WIND
        Toolkit dataset (https://www.nrel.gov/grid/wind-toolkit.html) for the
        specified lat/long coordinate in the continental US. The wind data
        are obtained from the WIND Toolkit dataset using the HSDS service (see
        https://github.com/NREL/hsds-examples). The wind data returned is
        obtained from the nearest 2km x 2km grid point to the input
        coordinate and is limited to the years 2007-2013. The wind parameters
        are binned using the specified wind parameter bin center values and the
        corresponding frequencies of occrance are calculated.

        Requires h5pyd package, which can be installed using:
            pip install --user git+http://github.com/HDFGroup/h5pyd.git

        Then, make a configuration file at ~/.hscfg containing:

            hs_endpoint = https://developer.nrel.gov/api/hsds

            hs_username = None

            hs_password = None

            hs_api_key = 3K3JQbjZmWctY0xmIfSYvYgtIcM3CN0cb1Y2w9bf

        The example API key above is for demonstation and is
        rate-limited per IP. To get your own API key, visit
        https://developer.nrel.gov/signup/.

        More information can be found at: https://github.com/NREL/hsds-examples.

        Args:
            lat (float): Latitude in degrees.
            lon (float): Longitude in degrees.
            ht (int, optional): The height above ground where wind
                information is obtained (m). Defaults to 100.
            wd (np.array, optional): Wind direction bin centers (deg). Defaults
                to np.arange(0, 360, 5.).
            ws (np.array, optional): Wind speed bin centers (m/s). Defaults to
                np.arange(0, 26, 1.).
            include_ti (bool, optional): Determines whether turbulence
                intensity is included as an additional parameter. If True, TI
                is added as an additional wind rose variable, estimated based
                on the Obukhov length from WIND Toolkit. Defaults to False.
            limit_month (list, optional): List of ints of month(s) (e.g., 1, 2,
                3...) to consider when calculating the wind condition
                frequencies. If none are specified, all months will be used.
                Defaults to None.
            limit_hour (list, optional): List of ints of hour(s) (e.g., 0, 1,
                ... 23) to consider when calculating the wind condition
                frequencies. If none are specified, all hours will be used.
                Defaults to None.
            st_date (str, optional): The start date to consider when creating
                the wind rose, formatted as 'MM-DD-YYYY'. If not specified data
                beginning in 2007 will be used. Defaults to None.
            en_date (str, optional): The end date to consider when creating
                the wind rose, formatted as 'MM-DD-YYYY'. If not specified data
                through 2013 will be used. Defaults to None.

        Returns:
            pandas.DataFrame: Wind rose DataFrame containing at least the
            following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                  wind conditions in the other columns.
        """
        # Check inputs

        # Array of hub height data avaliable on Toolkit
        h_range = [10, 40, 60, 80, 100, 120, 140, 160, 200]

        if st_date is not None:
            if dateutil.parser.parse(st_date) > dateutil.parser.parse(
                "12-13-2013 23:00"
            ):
                print(
                    "Error, invalid date range. Valid range: 01-01-2007 - "
                    + "12/31/2013"
                )
                return None

        if en_date is not None:
            if dateutil.parser.parse(en_date) < dateutil.parser.parse(
                "01-01-2007 00:00"
            ):
                print(
                    "Error, invalid date range. Valid range: 01-01-2007 - "
                    + "12/31/2013"
                )
                return None

        if h_range[0] > ht:
            print(
                "Error, height is not in the range of avaliable "
                + "WindToolKit data. Minimum height = 10m"
            )
            return None

        if h_range[-1] < ht:
            print(
                "Error, height is not in the range of avaliable "
                + "WindToolKit data. Maxiumum height = 200m"
            )
            return None

        # Load wind speeds and directions from WimdToolkit

        # Case for turbine height (ht) matching discrete avaliable height
        # (h_range)
        if ht in h_range:

            d = self.load_wind_toolkit_hsds(
                lat,
                lon,
                ht,
                include_ti=include_ti,
                limit_month=limit_month,
                limit_hour=limit_hour,
                st_date=st_date,
                en_date=en_date,
            )

            ws_new = d["ws"]
            wd_new = d["wd"]
            if include_ti:
                ti_new = d["ti"]

        # Case for ht not matching discete height
        else:
            h_range_up = next(x[0] for x in enumerate(h_range) if x[1] > ht)
            h_range_low = h_range_up - 1
            h_up = h_range[h_range_up]
            h_low = h_range[h_range_low]

            # Load data for boundary cases of ht
            d_low = self.load_wind_toolkit_hsds(
                lat,
                lon,
                h_low,
                include_ti=include_ti,
                limit_month=limit_month,
                limit_hour=limit_hour,
                st_date=st_date,
                en_date=en_date,
            )

            d_up = self.load_wind_toolkit_hsds(
                lat,
                lon,
                h_up,
                include_ti=include_ti,
                limit_month=limit_month,
                limit_hour=limit_hour,
                st_date=st_date,
                en_date=en_date,
            )

            # Wind Speed interpolation
            ws_low = d_low["ws"]
            ws_high = d_up["ws"]

            ws_new = np.array(ws_low) * (
                1 - ((ht - h_low) / (h_up - h_low))
            ) + np.array(ws_high) * ((ht - h_low) / (h_up - h_low))

            # Wind Direction interpolation using Circular Mean method
            wd_low = d_low["wd"]
            wd_high = d_up["wd"]

            sin0 = np.sin(np.array(wd_low) * (np.pi / 180))
            cos0 = np.cos(np.array(wd_low) * (np.pi / 180))
            sin1 = np.sin(np.array(wd_high) * (np.pi / 180))
            cos1 = np.cos(np.array(wd_high) * (np.pi / 180))

            sin_wd = sin0 * (1 - ((ht - h_low) / (h_up - h_low))) + sin1 * (
                (ht - h_low) / (h_up - h_low)
            )
            cos_wd = cos0 * (1 - ((ht - h_low) / (h_up - h_low))) + cos1 * (
                (ht - h_low) / (h_up - h_low)
            )

            # Interpolated wind direction
            wd_new = 180 / np.pi * np.arctan2(sin_wd, cos_wd)

            # TI is independent of height
            if include_ti:
                ti_new = d_up["ti"]

        # Create a dataframe named df
        if include_ti:
            df = pd.DataFrame({"ws": ws_new, "wd": wd_new, "ti": ti_new})
        else:
            df = pd.DataFrame({"ws": ws_new, "wd": wd_new})

        # Start by simply round and wrapping the wind direction and wind speed
        # columns
        df["wd"] = geo.wrap_360(df.wd.round())
        df["ws"] = df.ws.round()

        # Now group up
        df["freq_val"] = 1.0
        df = df.groupby([c for c in df.columns if c != "freq_val"]).sum()
        df["freq_val"] = df.freq_val.astype(float) / df.freq_val.sum()
        df = df.reset_index()

        # Save the df at this point
        self.df = df

        # Resample onto the provided wind speed and wind direction binnings
        self.internal_resample_wind_speed(ws=ws)
        self.internal_resample_wind_direction(wd=wd)

        return self.df

    def load_wind_toolkit_hsds(
        self,
        lat,
        lon,
        ht=100,
        include_ti=False,
        limit_month=None,
        limit_hour=None,
        st_date=None,
        en_date=None,
    ):
        """
        This method returns a pandas DataFrame containing hourly wind speed,
        wind direction, and optionally estimated turbulence intensity data
        using wind data from the WIND Toolkit dataset
        (https://www.nrel.gov/grid/wind-toolkit.html) for the specified
        lat/long coordinate in the continental US. The wind data are obtained
        from the WIND Toolkit dataset using the HSDS service
        (see https://github.com/NREL/hsds-examples). The wind data returned is
        obtained from the nearest 2km x 2km grid point to the input coordinate
        and is limited to the years 2007-2013.

        TODO: make private method?

        Args:
            lat (float): Latitude in degrees.
            lon (float): Longitude in degrees
            ht (int, optional): The height above ground where wind
                information is obtained (m). Defaults to 100.
            include_ti (bool, optional): Determines whether turbulence
                intensity is included as an additional parameter. If True, TI
                is added as an additional wind rose variable, estimated based
                on the Obukhov length from WIND Toolkit. Defaults to False.
            limit_month (list, optional): List of ints of month(s) (e.g., 1, 2,
                3...) to consider when calculating the wind condition
                frequencies. If none are specified, all months will be used.
                Defaults to None.
            limit_hour (list, optional): List of ints of hour(s) (e.g., 0, 1,
                ... 23) to consider when calculating the wind condition
                frequencies. If none are specified, all hours will be used.
                Defaults to None.
            st_date (str, optional): The start date to consider, formatted as
                'MM-DD-YYYY'. If not specified data beginning in 2007 will be
                used. Defaults to None.
            en_date (str, optional): The end date to consider, formatted as
                'MM-DD-YYYY'. If not specified data through 2013 will be used.
                Defaults to None.

        Returns:
            pandas.DataFrame: DataFrame containing the following columns(abd
            optionally turbulence intensity) with hourly data from WIND Toolkit:

                - **wd** (*float*) - Raw wind direction data (deg).
                - **ws** (*float*) - Raw wind speed data (m/s).
        """
        import h5pyd

        # Open the wind data "file"
        # server endpoint, username, password is found via a config file
        f = h5pyd.File("/nrel/wtk-us.h5", "r")

        # assign wind direction, wind speed, optional ti, and time datasets for
        # the desired height
        wd_dset = f["winddirection_" + str(ht) + "m"]
        ws_dset = f["windspeed_" + str(ht) + "m"]
        if include_ti:
            obkv_dset = f["inversemoninobukhovlength_2m"]
        dt = f["datetime"]
        dt = pd.DataFrame({"datetime": dt[:]}, index=range(0, dt.shape[0]))
        dt["datetime"] = dt["datetime"].apply(dateutil.parser.parse)

        # find dataset indices from lat/long
        Location_idx = self.indices_for_coord(f, lat, lon)

        # check if in bounds
        if (
            (Location_idx[0] < 0)
            | (Location_idx[0] >= wd_dset.shape[1])
            | (Location_idx[1] < 0)
            | (Location_idx[1] >= wd_dset.shape[2])
        ):
            print(
                "Error, coordinates out of bounds. WIND Toolkit database "
                + "covers the continental United States."
            )
            return None

        # create dataframe with wind direction and wind speed
        df = pd.DataFrame()
        df["wd"] = wd_dset[:, Location_idx[0], Location_idx[1]]
        df["ws"] = ws_dset[:, Location_idx[0], Location_idx[1]]
        if include_ti:
            L = self.obkv_dset_to_L(obkv_dset, Location_idx)
            ti = self.ti_calculator_IU2(L)
            df["ti"] = ti
        df["datetime"] = dt["datetime"]

        # limit dates if start and end dates are provided
        if st_date is not None:
            df = df[df.datetime >= st_date]

        if en_date is not None:
            df = df[df.datetime < en_date]

        # limit to certain months if specified
        if limit_month is not None:
            df["month"] = df["datetime"].map(lambda x: x.month)
            df = df[df.month.isin(limit_month)]
        if limit_hour is not None:
            df["hour"] = df["datetime"].map(lambda x: x.hour)
            df = df[df.hour.isin(limit_hour)]
        if include_ti:
            df = df[["wd", "ws", "ti"]]
        else:
            df = df[["wd", "ws"]]

        return df

    def obkv_dset_to_L(self, obkv_dset, Location_idx):
        """
        This function returns an array containing hourly Obukhov lengths from
        the WIND Toolkit dataset for the specified Lat/Lon coordinate indices.

        Args:
            obkv_dset (np.ndarray): Dataset for Obukhov lengths from WIND
                Toolkit.
            Location_idx (tuple): A tuple containing the Lat/Lon coordinate
                indices of interest in the Obukhov length dataset.

        Returns:
            np.array: An array containing Obukhov lengths for each time index
            in the Wind Toolkit dataset (m).
        """
        linv = obkv_dset[:, Location_idx[0], Location_idx[1]]
        # avoid divide by zero
        linv[linv == 0.0] = 0.0003
        L = 1 / linv
        return L

    def ti_calculator_IU2(self, L):
        """
        This function estimates the turbulence intensity corresponding to each
        Obukhov length value in the input list using the relationship between
        Obukhov length bins and TI given in the I_U2SODAR column in Table 2 of
        :cite:`wr-wharton2010assessing`.

        Args:
            L (iterable): A list of Obukhov Length values (m).

        Returns:
            list: A list of turbulence intensity values expressed as fractions.
        """
        ti_set = []
        for i in L:
            # Strongly Stable
            if 0 < i < 100:
                TI = 0.04  # paper says < 8%, so using 4%
            # Stable
            elif 100 < i < 600:
                TI = 0.09
            # Neutral
            elif abs(i) > 600:
                TI = 0.115
            # Convective
            elif -600 < i < -50:
                TI = 0.165
            # Strongly Convective
            elif -50 < i < 0:
                # no upper bound given, so using the lowest
                # value from the paper for this stability bin
                TI = 0.2
            ti_set.append(TI)
        return ti_set

    def indices_for_coord(self, f, lat_index, lon_index):
        """
        This method finds the nearest x/y indices of the WIND Toolkit dataset
        for a given lat/lon coordinate in the continental US. Rather than
        fetching the entire coordinates database, which is 500+ MB, this uses
        the Proj4 library to find a nearby point and then converts to x/y
        indices.

        **Note**: This method is obtained directly from:
        https://github.com/NREL/hsds-examples/blob/master/notebooks/01_WTK_introduction.ipynb,
        where it is called "indicesForCoord."

        Args:
            f (h5pyd.File): A HDF5 "file" used to access the WIND Toolkit data.
            lat_index (float): Latitude coordinate for which dataset indices
                are to be found (degrees).
            lon_index (float): Longitude coordinate for which dataset indices
                are to be found (degrees).

        Returns:
            tuple: A tuple containing the Lat/Lon coordinate indices of
            interest in the WIND Toolkit dataset.
        """
        dset_coords = f["coordinates"]
        projstring = """+proj=lcc +lat_1=30 +lat_2=60
                        +lat_0=38.47240422490422 +lon_0=-96.0
                        +x_0=0 +y_0=0 +ellps=sphere
                        +units=m +no_defs """
        projectLcc = Proj(projstring)
        origin_ll = reversed(dset_coords[0][0])  # Grab origin directly from database
        origin = projectLcc(*origin_ll)

        coords = (lon_index, lat_index)
        coords = projectLcc(*coords)
        delta = np.subtract(coords, origin)
        ij = [int(round(x / 2000)) for x in delta]
        return tuple(reversed(ij))

    def plot_wind_speed_all(self, ax=None):
        """
        This method plots the wind speed frequency distribution of the WindRose
        object averaged across all wind directions. If no axis is provided, a
        new one is created.

        Args:
            ax (:py:class:`matplotlib.pyplot.axes`, optional): Figure axes on
                which data should be plotted. Defaults to None.
        """
        if ax is None:
            _, ax = plt.subplots()

        df_plot = self.df.groupby("ws").sum()
        ax.plot(self.ws, df_plot.freq_val)

    def plot_wind_speed_by_direction(self, dirs, ax=None):
        """
        This method plots the wind speed frequency distribution of the WindRose
        object for each specified wind direction bin center. The wind
        directions are resampled using the specified bin centers and the
        frequencies of occurance of the wind conditions are modified
        accordingly. If no axis is provided, a new one is created.

        Args:
            dirs (np.array): A list of wind direction bin centers for which
                wind speed distributions are plotted (deg).
            ax (:py:class:`matplotlib.pyplot.axes`, optional): Figure axes on
                which data should be plotted. Defaults to None.
        """
        # Get a downsampled frame
        df_plot = self.resample_wind_direction(self.df, wd=dirs)

        if ax is None:
            _, ax = plt.subplots()

        for wd in dirs:
            df_plot_sub = df_plot[df_plot.wd == wd]
            ax.plot(df_plot_sub.ws, df_plot_sub["freq_val"], label=wd)
        ax.legend()

    def plot_wind_rose(
        self,
        ax=None,
        color_map="viridis_r",
        ws_right_edges=np.array([5, 10, 15, 20, 25]),
        wd_bins=np.arange(0, 360, 15.0),
        legend_kwargs={},
    ):
        """
        This method creates a wind rose plot showing the frequency of occurance
        of the specified wind direction and wind speed bins. If no axis is
        provided, a new one is created.

        **Note**: Based on code provided by Patrick Murphy from the University
        of Colorado Boulder.

        Args:
            ax (:py:class:`matplotlib.pyplot.axes`, optional): The figure axes
                on which the wind rose is plotted. Defaults to None.
            color_map (str, optional): Colormap to use. Defaults to 'viridis_r'.
            ws_right_edges (np.array, optional): The upper bounds of the wind
                speed bins (m/s). The first bin begins at 0. Defaults to
                np.array([5, 10, 15, 20, 25]).
            wd_bins (np.array, optional): The wind direction bin centers used
                for plotting (deg). Defaults to np.arange(0, 360, 15.).
            legend_kwargs (dict, optional): Keyword arguments to be passed to
                ax.legend().

        Returns:
            :py:class:`matplotlib.pyplot.axes`: A figure axes object containing
            the plotted wind rose.
        """
        # Resample data onto bins
        df_plot = self.resample_wind_direction(self.df, wd=wd_bins)

        # Make labels for wind speed based on edges
        ws_step = ws_right_edges[1] - ws_right_edges[0]
        ws_labels = ["%d-%d m/s" % (w - ws_step, w) for w in ws_right_edges]

        # Grab the wd_step
        wd_step = wd_bins[1] - wd_bins[0]

        # Set up figure
        if ax is None:
            _, ax = plt.subplots(subplot_kw=dict(polar=True))

        # Get a color array
        color_array = cm.get_cmap(color_map, len(ws_right_edges))

        for wd_idx, wd in enumerate(wd_bins):
            rects = list()
            df_plot_sub = df_plot[df_plot.wd == wd]
            for ws_idx, ws in enumerate(ws_right_edges[::-1]):
                plot_val = df_plot_sub[
                    df_plot_sub.ws <= ws
                ].freq_val.sum()  # Get the sum of frequency up to this wind speed
                rects.append(
                    ax.bar(
                        np.radians(wd),
                        plot_val,
                        width=0.9 * np.radians(wd_step),
                        color=color_array(ws_idx),
                        edgecolor="k",
                    )
                )
            # break

        # Configure the plot
        ax.legend(reversed(rects), ws_labels, **legend_kwargs)
        ax.set_theta_direction(-1)
        ax.set_theta_offset(np.pi / 2.0)
        ax.set_theta_zero_location("N")
        ax.set_xticklabels(["N", "NE", "E", "SE", "S", "SW", "W", "NW"])

        return ax

    def plot_wind_rose_ti(
        self,
        ax=None,
        color_map="viridis_r",
        ti_right_edges=np.array([0.06, 0.1, 0.14, 0.18, 0.22]),
        wd_bins=np.arange(0, 360, 15.0),
    ):
        """
        This method creates a wind rose plot showing the frequency of occurance
        of the specified wind direction and turbulence intensity bins. This
        requires turbulence intensity to already be included as a parameter in
        the wind rose. If no axis is provided,a new one is created.

        **Note**: Based on code provided by Patrick Murphy from the University
        of Colorado Boulder.

        Args:
            ax (:py:class:`matplotlib.pyplot.axes`, optional): The figure axes
                on which the wind rose is plotted. Defaults to None.
            color_map (str, optional): Colormap to use. Defaults to 'viridis_r'.
            ti_right_edges (np.array, optional): The upper bounds of the
                turbulence intensity bins. The first bin begins at 0. Defaults
                to np.array([0.06, 0.1, 0.14, 0.18,0.22]).
            wd_bins (np.array, optional): The wind direction bin centers used
                for plotting (deg). Defaults to np.arange(0, 360, 15.).

        Returns:
            :py:class:`matplotlib.pyplot.axes`: A figure axes object containing
            the plotted wind rose.
        """

        # Resample data onto bins
        df_plot = self.resample_wind_direction(self.df, wd=wd_bins)

        # Make labels for TI based on edges
        ti_step = ti_right_edges[1] - ti_right_edges[0]
        ti_labels = ["%.2f-%.2f " % (w - ti_step, w) for w in ti_right_edges]

        # Grab the wd_step
        wd_step = wd_bins[1] - wd_bins[0]

        # Set up figure
        if ax is None:
            _, ax = plt.subplots(subplot_kw=dict(polar=True))

        # Get a color array
        color_array = cm.get_cmap(color_map, len(ti_right_edges))

        for wd_idx, wd in enumerate(wd_bins):
            rects = list()
            df_plot_sub = df_plot[df_plot.wd == wd]
            for ti_idx, ti in enumerate(ti_right_edges[::-1]):
                plot_val = df_plot_sub[
                    df_plot_sub.ti <= ti
                ].freq_val.sum()  # Get the sum of frequency up to this wind speed
                rects.append(
                    ax.bar(
                        np.radians(wd),
                        plot_val,
                        width=0.9 * np.radians(wd_step),
                        color=color_array(ti_idx),
                        edgecolor="k",
                    )
                )

        # Configure the plot
        ax.legend(reversed(rects), ti_labels, loc="lower right", title="TI")
        ax.set_theta_direction(-1)
        ax.set_theta_offset(np.pi / 2.0)
        ax.set_theta_zero_location("N")
        ax.set_xticklabels(["N", "NE", "E", "SE", "S", "SW", "W", "NW"])

        return ax

    def plot_ti_ws(self, ax=None, ws_bins=np.arange(0, 26, 1.0)):
        """
        This method plots the wind speed frequency distribution of the WindRose
        object for each turbulence intensity bin. The wind speeds are resampled
        using the specified bin centers and the frequencies of occurance of the
        wind conditions are modified accordingly. This method assumes there are
        five TI bins. If no axis is provided, a new one is created.

        Args:
            ax (:py:class:`matplotlib.pyplot.axes`, optional): Figure axes on
                which data should be plotted. Defaults to None.
            ws_bins (np.array, optional): A list of wind speed bin centers on
                which the wind speeds are resampled before plotting (m/s).
                Defaults to np.arange(0, 26, 1.).

        Returns:
            :py:class:`matplotlib.pyplot.axes`: A figure axes object containing
            the plotted wind speed distributions.
        """

        # Resample data onto bins
        df_plot = self.resample_wind_speed(self.df, ws=ws_bins)

        df_plot = df_plot.groupby(["ws", "ti"]).sum()
        df_plot = df_plot.reset_index()

        if ax is None:
            _, ax = plt.subplots(figsize=(10, 7))

        tis = df_plot["ti"].drop_duplicates()
        margin_bottom = np.zeros(len(df_plot["ws"].drop_duplicates()))
        colors = ["#1e5631", "#a4de02", "#76ba1b", "#4c9a2a", "#acdf87"]

        for num, ti in enumerate(tis):
            values = list(df_plot[df_plot["ti"] == ti].loc[:, "freq_val"])

            df_plot[df_plot["ti"] == ti].plot.bar(
                x="ws",
                y="freq_val",
                ax=ax,
                bottom=margin_bottom,
                color=colors[num],
                label=ti,
            )

            margin_bottom += values

        plt.title("Turbulence Intensity Frequencies as Function of Wind Speed")
        plt.xlabel("Wind Speed (m/s)")
        plt.ylabel("Frequency")

        return ax

    def export_for_floris_opt(self):
        """
        This method returns a list of tuples of at least wind speed, wind
        direction, and frequency of occurance, which can be used to help loop
        through different wind conditions for Floris power calculations.

        Returns:
            list: A list of tuples containing all combinations of wind
            parameters and frequencies of occurance in the WindRose object's
            wind rose DataFrame values.
        """
        # Return a list of tuples, where each tuple is (ws,wd,freq)
        return [tuple(x) for x in self.df.values]



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize

from .layout import LayoutOptimization


class PowerDensityOptimization(LayoutOptimization):
    """
    PowerDensityOptimization is a subclass of the
    :py:class:`~.tools.optimization.scipy.layout.LayoutOptimization` class
    that performs power density optimization.
    """

    def __init__(
        self,
        fi,
        boundaries,
        wd,
        ws,
        freq,
        AEP_initial,
        yawbnds=None,
        x0=None,
        bnds=None,
        min_dist=None,
        opt_method="SLSQP",
        opt_options=None,
    ):
        """
        Instantiate PowerDensityOptimization object with a FlorisInterface
        object and assigns parameter values.

        Args:
            fi (:py:class:`floris.tools.floris_interface.FlorisInterface`):
                Interface used to interact with the Floris object.
            boundaries (iterable(float, float)): Pairs of x- and y-coordinates
                that represent the boundary's vertices (m).
            wd (np.array): An array of wind directions (deg).
            ws (np.array): An array of wind speeds (m/s).
            freq (np.array): An array of the frequencies of occurance
                correponding to each pair of wind direction and wind speed
                values.
            AEP_initial (float): The initial Annual Energy
                Production used for normalization in the optimization (Wh)
                (TODO: Is Watt-hours the correct unit?).
            yawbnds: TODO: This parameter isn't used. Remove it?
            x0 (iterable, optional): The initial turbine locations,
                ordered by x-coordinate and then y-coordiante
                (ie. [x1, x2, ..., xn, y1, y2, ..., yn]) (m). If none are
                provided, x0 initializes to the current turbine locations.
                Defaults to None.
            bnds (iterable, optional): Bounds for the optimization
                variables (pairs of min/max values for each variable (m)). If
                none are specified, they are set to (0, 1) for each turbine.
                Defaults to None. TODO: Explain significance of (0, 1).
            min_dist (float, optional): The minimum distance to be
                maintained between turbines during the optimization (m). If not
                specified, initializes to 4 rotor diameters. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to 'SLSQP'.
            opt_options (dict, optional): Optimization options used by
                scipy.optimize.minize. If none are specified, they are set t
                {'maxiter': 100, 'disp': True, 'iprint': 2, 'ftol': 1e-9}.
                Defaults to None.
        """
        super().__init__(
            fi,
            boundaries,
            wd,
            ws,
            freq,
            AEP_initial,
            x0=x0,
            bnds=bnds,
            min_dist=min_dist,
            opt_method=opt_method,
            opt_options=opt_options,
        )
        self.epsilon = np.finfo(float).eps
        self.counter = 0

        if opt_options is None:
            self.opt_options = {"maxiter": 100, "disp": True, "iprint": 2, "ftol": 1e-9}

    def _generate_constraints(self):
        # grad_constraint1 = grad(self._space_constraint)
        # grad_constraint2 = grad(self._distance_from_boundaries)

        tmp1 = {
            "type": "ineq",
            "fun": lambda x, *args: self._space_constraint(x, self.min_dist),
            "args": (self.min_dist,),
        }
        tmp2 = {
            "type": "ineq",
            "fun": lambda x, *args: self._distance_from_boundaries(
                x, self.boundaries_norm
            ),
            "args": (self.boundaries_norm,),
        }
        tmp3 = {"type": "ineq", "fun": lambda x, *args: self._AEP_constraint(x)}

        self.cons = [tmp1, tmp2, tmp3]

    def _set_opt_bounds(self):
        self.bnds = [
            (0.0, 1.0) for _ in range(2 * self.nturbs + self.nturbs * len(self.wd))
        ]

    def _change_coordinates(self, locsx, locsy):
        # Parse the layout coordinates
        layout_array = [locsx, locsy]

        # Update the turbine map in floris
        self.fi.reinitialize_flow_field(layout_array=layout_array)

    def _powDens_opt(self, optVars):
        locsx = optVars[0 : self.nturbs]
        locsy = optVars[self.nturbs : 2 * self.nturbs]

        locsx_unnorm = [
            self._unnorm(valx, self.bndx_min, self.bndx_max) for valx in locsx
        ]
        locsy_unnorm = [
            self._unnorm(valy, self.bndy_min, self.bndy_max) for valy in locsy
        ]

        turb_controls = [
            optVars[
                2 * self.nturbs + i * self.nturbs : 3 * self.nturbs + i * self.nturbs
            ]
            for i in range(len(self.wd))
        ]

        turb_controls_unnorm = [
            self._unnorm(yaw, self.yaw_min, self.yaw_max) for yaw in turb_controls
        ]

        self._change_coordinates(locsx_unnorm, locsy_unnorm)
        opt_area = self.find_layout_area(locsx_unnorm + locsy_unnorm)

        AEP_sum = 0.0

        for i in range(len(self.wd)):
            for j, turbine in enumerate(self.fi.floris.farm.turbine_map.turbines):
                turbine.yaw_angle = turb_controls_unnorm[i][j]

            AEP_sum = AEP_sum + self._AEP_single_wd(
                self.wd[i], self.ws[i], self.freq[i]
            )

        # print('AEP ratio: ', AEP_sum/self.AEP_initial)

        return -1 * AEP_sum / self.AEP_initial * self.initial_area / opt_area

    def _AEP_constraint(self, optVars):
        locsx = optVars[0 : self.nturbs]
        locsy = optVars[self.nturbs : 2 * self.nturbs]

        locsx_unnorm = [
            self._unnorm(valx, self.bndx_min, self.bndx_max) for valx in locsx
        ]
        locsy_unnorm = [
            self._unnorm(valy, self.bndy_min, self.bndy_max) for valy in locsy
        ]

        turb_controls = [
            optVars[
                2 * self.nturbs + i * self.nturbs : 3 * self.nturbs + i * self.nturbs
            ]
            for i in range(len(self.wd))
        ]

        turb_controls_unnorm = [
            self._unnorm(yaw, self.yaw_min, self.yaw_max) for yaw in turb_controls
        ]

        self._change_coordinates(locsx_unnorm, locsy_unnorm)

        AEP_sum = 0.0

        for i in range(len(self.wd)):
            for j, turbine in enumerate(self.fi.floris.farm.turbine_map.turbines):
                turbine.yaw_angle = turb_controls_unnorm[i][j]

            AEP_sum = AEP_sum + self._AEP_single_wd(
                self.wd[i], self.ws[i], self.freq[i]
            )

        return AEP_sum / self.AEP_initial - 1.0

    def _optimize(self):
        self.residual_plant = minimize(
            self._powDens_opt,
            self.x0,
            method=self.opt_method,
            bounds=self.bnds,
            constraints=self.cons,
            options=self.opt_options,
        )

        opt_results = self.residual_plant.x

        return opt_results

    def optimize(self):
        """
        This method finds the optimized layout of wind turbines for power
        production given the provided frequencies of occurance of wind
        conditions (wind speed, direction).

        TODO: update the doc

        Returns:
            iterable: A list of the optimized x, y locations of each
            turbine (m).
        """
        print("=====================================================")
        print("Optimizing turbine layout...")
        print("Number of parameters to optimize = ", len(self.x0))
        print("=====================================================")

        opt_locs_norm = self._optimize()

        print("Optimization complete.")

        opt_locs = [
            [
                self._unnorm(valx, self.bndx_min, self.bndx_max)
                for valx in opt_locs_norm[0 : self.nturbs]
            ],
            [
                self._unnorm(valy, self.bndy_min, self.bndy_max)
                for valy in opt_locs_norm[self.nturbs : 2 * self.nturbs]
            ],
        ]

        return opt_locs

    def reinitialize_opt(
        self,
        boundaries=None,
        yawbnds=None,
        wd=None,
        ws=None,
        freq=None,
        AEP_initial=None,
        x0=None,
        bnds=None,
        min_dist=None,
        opt_method=None,
        opt_options=None,
    ):
        """
        This method reinitializes any optimization parameters that are
        specified. Otherwise, the current parameter values are kept.

        Args:
            boundaries (iterable(float, float)): Pairs of x- and y-coordinates
                that represent the boundary's vertices (m).
            yawbnds (iterable): A list of the min. and max. yaw offset that is
                allowed during the optimization (deg). If none are specified,
                initialized to (0, 25.0). Defaults to None.
            wd (np.array): An array of wind directions (deg). Defaults to None.
            ws (np.array): An array of wind speeds (m/s). Defaults to None.
            freq (np.array): An array of the frequencies of occurance
                correponding to each pair of wind direction and wind speed
                values. Defaults to None.
            AEP_initial (float): The initial Annual Energy
                Production used for normalization in the optimization (Wh)
                (TODO: Is Watt-hours the correct unit?). If not specified,
                initializes to the AEP of the current Floris object. Defaults
                to None.
            x0 (iterable, optional): The initial turbine locations,
                ordered by x-coordinate and then y-coordiante
                (ie. [x1, x2, ..., xn, y1, y2, ..., yn]) (m). If none are
                provided, x0 initializes to the current turbine locations.
                Defaults to None.
            bnds (iterable, optional): Bounds for the optimization
                variables (pairs of min/max values for each variable (m)). If
                none are specified, they are set to (0, 1) for each turbine.
                Defaults to None.
            min_dist (float, optional): The minimum distance to be
                maintained between turbines during the optimization (m). If not
                specified, initializes to 4 rotor diameters. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to None.
            opt_options (dict, optional): Optimization options used by
                scipy.optimize.minize. Defaults to None.
        """
        if boundaries is not None:
            self.boundaries = boundaries
            self.bndx_min = np.min([val[0] for val in boundaries])
            self.bndy_min = np.min([val[1] for val in boundaries])
            self.bndx_max = np.max([val[0] for val in boundaries])
            self.bndy_max = np.max([val[1] for val in boundaries])
            self.boundaries_norm = [
                [
                    self._norm(val[0], self.bndx_min, self.bndx_max),
                    self._norm(val[1], self.bndy_min, self.bndy_max),
                ]
                for val in self.boundaries
            ]
        if yawbnds is not None:
            self.yaw_min = yawbnds[0]
            self.yaw_max = yawbnds[1]
        else:
            self.yaw_min = 0.0
            self.yaw_max = 25.0
        if wd is not None:
            self.wd = wd
        if ws is not None:
            self.ws = ws
        if freq is not None:
            self.freq = freq
        if AEP_initial is not None:
            self.AEP_initial = AEP_initial
        else:
            self.AEP_initial = self.fi.get_farm_AEP(self.wd, self.ws, self.freq)
        if x0 is not None:
            self.x0 = x0
        else:
            self.x0 = (
                [
                    self._norm(coord.x1, self.bndx_min, self.bndx_max)
                    for coord in self.fi.floris.farm.turbine_map.coords
                ]
                + [
                    self._norm(coord.x2, self.bndy_min, self.bndy_max)
                    for coord in self.fi.floris.farm.turbine_map.coords
                ]
                + [self._norm(5.0, self.yaw_min, self.yaw_max)]
                * len(self.wd)
                * self.nturbs
            )
        if bnds is not None:
            self.bnds = bnds
        else:
            self._set_opt_bounds()
        if min_dist is not None:
            self.min_dist = min_dist
        else:
            self.min_dist = 4 * self.fi.floris.farm.turbines[0].rotor_diameter
        if opt_method is not None:
            self.opt_method = opt_method
        if opt_options is not None:
            self.opt_options = opt_options

        self.layout_x_orig = [
            coord.x1 for coord in self.fi.floris.farm.turbine_map.coords
        ]
        self.layout_y_orig = [
            coord.x2 for coord in self.fi.floris.farm.turbine_map.coords
        ]

        self._generate_constraints()

        self.initial_area = self.find_layout_area(
            self.layout_x_orig + self.layout_y_orig
        )

    def find_layout_area(self, locs):
        """
        This method returns the area occupied by the wind farm.

        Args:
            locs (iterable): A list of the turbine coordinates, organized as
                [x1, x2, ..., xn, y1, y2, ..., yn] (m).

        Returns:
            float: The area occupied by the wind farm (m^2).
        """
        locsx = locs[0 : self.nturbs]
        locsy = locs[self.nturbs :]

        points = zip(locsx, locsy)
        points = np.array(list(points))

        hull = self.convex_hull(points)

        area = self.polygon_area(
            np.array([val[0] for val in hull]), np.array([val[1] for val in hull])
        )

        return area

    def convex_hull(self, points):
        """
        Finds the vertices that describe the convex hull shape given the input
        coordinates.

        Args:
            points (iterable((float, float))): Coordinates of interest.

        Returns:
            list: Vertices describing convex hull shape.
        """
        # find two hull points, U, V, and split to left and right search
        u = min(points, key=lambda p: p[0])
        v = max(points, key=lambda p: p[0])
        left, right = self.split(u, v, points), self.split(v, u, points)

        # find convex hull on each side
        return [v] + self.extend(u, v, left) + [u] + self.extend(v, u, right) + [v]

    def polygon_area(self, x, y):
        """
        Calculates the area of a polygon defined by its (x, y) vertices.

        Args:
            x (iterable(float)): X-coordinates of polygon vertices.
            y (iterable(float)): Y-coordinates of polygon vertices.

        Returns:
            float: Area of polygon.
        """
        # coordinate shift
        x_ = x - x.mean()
        y_ = y - y.mean()

        correction = x_[-1] * y_[0] - y_[-1] * x_[0]
        main_area = np.dot(x_[:-1], y_[1:]) - np.dot(y_[:-1], x_[1:])
        return 0.5 * np.abs(main_area + correction)

    def split(self, u, v, points):
        # TODO: Provide description of this method.
        # return points on left side of UV
        return [p for p in points if np.cross(p - u, v - u) < 0]

    def extend(self, u, v, points):
        # TODO: Provide description of this method.
        if not points:
            return []

        # find furthest point W, and split search to WV, UW
        w = min(points, key=lambda p: np.cross(p - u, v - u))
        p1, p2 = self.split(w, v, points), self.split(u, w, points)
        return self.extend(w, v, p1) + [w] + self.extend(u, w, p2)

    def plot_opt_results(self):
        """
        This method plots the original and new locations of the turbines in a
        wind farm after layout optimization.
        """
        locsx_old = [
            self._unnorm(valx, self.bndx_min, self.bndx_max)
            for valx in self.x0[0 : self.nturbs]
        ]
        locsy_old = [
            self._unnorm(valy, self.bndy_min, self.bndy_max)
            for valy in self.x0[self.nturbs : 2 * self.nturbs]
        ]
        locsx = [
            self._unnorm(valx, self.bndx_min, self.bndx_max)
            for valx in self.residual_plant.x[0 : self.nturbs]
        ]
        locsy = [
            self._unnorm(valy, self.bndy_min, self.bndy_max)
            for valy in self.residual_plant.x[self.nturbs : 2 * self.nturbs]
        ]

        plt.figure(figsize=(9, 6))
        fontsize = 16
        plt.plot(locsx_old, locsy_old, "ob")
        plt.plot(locsx, locsy, "or")
        # plt.title('Layout Optimization Results', fontsize=fontsize)
        plt.xlabel("x (m)", fontsize=fontsize)
        plt.ylabel("y (m)", fontsize=fontsize)
        plt.axis("equal")
        plt.grid()
        plt.tick_params(which="both", labelsize=fontsize)
        plt.legend(
            ["Old locations", "New locations"],
            loc="lower center",
            bbox_to_anchor=(0.5, 1.01),
            ncol=2,
            fontsize=fontsize,
        )

        verts = self.boundaries
        for i in range(len(verts)):
            if i == len(verts) - 1:
                plt.plot([verts[i][0], verts[0][0]], [verts[i][1], verts[0][1]], "b")
            else:
                plt.plot(
                    [verts[i][0], verts[i + 1][0]], [verts[i][1], verts[i + 1][1]], "b"
                )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# from __future__ import annotations
from typing import Any, List

import attrs
from attrs import define, field
import numpy as np

from floris.type_dec import (
    floris_array_converter,
    NDArrayFloat
)
from floris.utilities import Vec3
from floris.simulation import BaseClass


@define
class Farm(BaseClass):
    """Farm is where wind power plants should be instantiated from a YAML configuration
    file. The Farm will create a heterogenous set of turbines that compose a windfarm,
    validate the inputs, and then create a vectorized representation of the the turbine
    data.

    Farm is the container class of the FLORIS package. It brings
    together all of the component objects after input (i.e., Turbine,
    Wake, FlowField) and packages everything into the appropriate data
    type. Farm should also be used as an entry point to probe objects
    for generating output.
    """

    layout_x: NDArrayFloat = field(converter=floris_array_converter)
    layout_y: NDArrayFloat = field(converter=floris_array_converter)

    yaw_angles: NDArrayFloat = field(init=False)
    coordinates: List[Vec3] = field(init=False)

    @layout_x.validator
    def check_x(self, instance: attrs.Attribute, value: Any) -> None:
        if len(value) != len(self.layout_y):
            raise ValueError("layout_x and layout_y must have the same number of entries.")

    @layout_y.validator
    def check_y(self, instance: attrs.Attribute, value: Any) -> None:
        if len(value) != len(self.layout_x):
            raise ValueError("layout_x and layout_y must have the same number of entries.")

    def initialize(self, sorted_indices):
        # Sort yaw angles from most upstream to most downstream wind turbine
        self.yaw_angles = np.take_along_axis(
            self.yaw_angles,
            sorted_indices[:, :, :, 0, 0],
            axis=2,
        )

    def construct_coordinates(self, reference_z: float):
        self.coordinates = np.array(
            [Vec3([x, y, reference_z]) for x, y in zip(self.layout_x, self.layout_y)]
        )

    def set_yaw_angles(self, n_wind_directions: int, n_wind_speeds: int):
        # TODO Is this just for initializing yaw angles to zero?
        self.yaw_angles = np.zeros((n_wind_directions, n_wind_speeds, self.n_turbines))

    def finalize(self, unsorted_indices):
        self.yaw_angles = np.take_along_axis(self.yaw_angles, unsorted_indices[:,:,:,0,0], axis=2)

    @property
    def n_turbines(self):
        return len(self.layout_x)



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from attrs import define
import numpy as np

from floris.simulation import BaseModel


@define
class MAX(BaseModel):
    """
    MAX uses the maximum wake velocity deficit to add to the
    base flow field. For more information, refer to
    :cite:`max-gunn2016limitations`.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: max-
    """

    model_string = "max"

    def prepare_function(self) -> dict:
        pass

    def function(self, wake_field: np.ndarray, velocity_field: np.ndarray):
        """
        Incorporates the velicty deficits into the base flow field by
        selecting the maximum of the two for each point.

        Args:
            u_field (np.array): The base flow field.
            u_wake (np.array): The wake to apply to the base flow field.

        Returns:
            np.array: The resulting flow field after applying the wake to the
                base.
        """
        return np.maximum(wake_field, velocity_field)



# Copyright 2021 NREL
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


from __future__ import annotations

import copy
from typing import Any, Tuple
from pathlib import Path
from itertools import repeat, product
from multiprocessing import cpu_count
from multiprocessing.pool import Pool

import numpy as np
import pandas as pd
import numpy.typing as npt
import matplotlib.pyplot as plt
from scipy.stats import norm
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator
from numpy.lib.arraysetops import unique

from floris.utilities import Vec3
from floris.type_dec import NDArrayFloat
from floris.simulation import Farm, Floris, FlowField, WakeModelManager, farm, floris, flow_field
from floris.logging_manager import LoggerBase
from floris.tools.cut_plane import get_plane_from_flow_data
# from floris.tools.flow_data import FlowData
from floris.simulation.turbine import Ct, power, axial_induction, average_velocity
from floris.tools.interface_utilities import get_params, set_params, show_params
from floris.tools.cut_plane import CutPlane, change_resolution, get_plane_from_flow_data

# from .visualization import visualize_cut_plane
# from .layout_functions import visualize_layout, build_turbine_loc


class FlorisInterface(LoggerBase):
    """
    FlorisInterface provides a high-level user interface to many of the
    underlying methods within the FLORIS framework. It is meant to act as a
    single entry-point for the majority of users, simplifying the calls to
    methods on objects within FLORIS.

    Args:
        configuration (:py:obj:`dict`): The Floris configuration dictarionary, JSON file,
            or YAML file. The configuration should have the following inputs specified.
                - **flow_field**: See `floris.simulation.flow_field.FlowField` for more details.
                - **farm**: See `floris.simulation.farm.Farm` for more details.
                - **turbine**: See `floris.simulation.turbine.Turbine` for more details.
                - **wake**: See `floris.simulation.wake.WakeManager` for more details.
                - **logging**: See `floris.simulation.floris.Floris` for more details.
    """

    def __init__(self, configuration: dict | str | Path, het_map=None):
        self.configuration = configuration

        if isinstance(self.configuration, (str, Path)):
            self.floris = Floris.from_file(self.configuration)

        elif isinstance(self.configuration, dict):
            self.floris = Floris.from_dict(self.configuration)

        else:
            raise TypeError("The Floris `configuration` must of type 'dict', 'str', or 'Path'.")

        # Store the heterogeneous map for use after reinitailization
        self.het_map = het_map
        # Assign the heterogeneous map to the flow field
        # Needed for a direct call to fi.calculate_wake without fi.reinitialize
        self.floris.flow_field.het_map = het_map

    def calculate_wake(
        self,
        yaw_angles: NDArrayFloat | list[float] | None = None,
        # no_wake: bool = False,
        # points: NDArrayFloat | list[float] | None = None,
        # track_n_upstream_wakes: bool = False,
    ) -> None:
        """
        Wrapper to the :py:meth:`~.Farm.set_yaw_angles` and
        :py:meth:`~.FlowField.calculate_wake` methods.

        Args:
            yaw_angles (NDArrayFloat | list[float] | None, optional): Turbine yaw angles.
                Defaults to None.
            no_wake: (bool, optional): When *True* updates the turbine
                quantities without calculating the wake or adding the
                wake to the flow field. Defaults to *False*.
            points: (NDArrayFloat | list[float] | None, optional): The x, y, and z
                coordinates at which the flow field velocity is to be recorded. Defaults
                to None.
            track_n_upstream_wakes (bool, optional): When *True*, will keep track of the
                number of upstream wakes a turbine is experiencing. Defaults to *False*.
        """
        # self.floris.flow_field.calculate_wake(
        #     no_wake=no_wake,
        #     points=points,
        #     track_n_upstream_wakes=track_n_upstream_wakes,
        # )

        # TODO decide where to handle this sign issue
        if yaw_angles is not None:
            self.floris.farm.yaw_angles = yaw_angles

        self.floris.steady_state_atmospheric_condition()

    def reinitialize(
        self,
        wind_speeds: list[float] | NDArrayFloat | None = None,
        wind_directions: list[float] | NDArrayFloat | None = None,
        # wind_layout: list[float] | NDArrayFloat | None = None,
        wind_shear: float | None = None,
        wind_veer: float | None = None,
        reference_wind_height: float | None = None,
        turbulence_intensity: float | None = None,
        # turbulence_kinetic_energy=None,
        air_density: float | None = None,
        # wake: WakeModelManager = None,
        layout: Tuple[list[float], list[float]] | Tuple[NDArrayFloat, NDArrayFloat] | None = None,
        # turbine_id: list[str] | None = None,
        # wtg_id: list[str] | None = None,
        # with_resolution: float | None = None,
        solver_settings: dict | None = None
    ):
        # Export the floris object recursively as a dictionary
        floris_dict = self.floris.as_dict()
        flow_field_dict = floris_dict["flow_field"]
        farm_dict = floris_dict["farm"]

        # Make the given changes

        ## FlowField
        if wind_speeds is not None:
            flow_field_dict["wind_speeds"] = wind_speeds
        if wind_directions is not None:
            flow_field_dict["wind_directions"] = wind_directions
        if wind_shear is not None:
            flow_field_dict["wind_shear"] = wind_shear
        if wind_veer is not None:
            flow_field_dict["wind_veer"] = wind_veer
        if reference_wind_height is not None:
            flow_field_dict["reference_wind_height"] = reference_wind_height
        if turbulence_intensity is not None:
            flow_field_dict["turbulence_intensity"] = turbulence_intensity
        if air_density is not None:
            flow_field_dict["air_density"] = air_density

        if layout is not None:
            farm_dict["layout_x"] = layout[0]
            farm_dict["layout_y"] = layout[1]

        ## Wake
        # if wake is not None:
        #     self.floris.wake = wake
        # if turbulence_intensity is not None:
        #     pass  # TODO: this should be in the code, but maybe got skipped?
        # if turbulence_kinetic_energy is not None:
        #     pass  # TODO: not needed until GCH
        if solver_settings is not None:
            floris_dict["solver"] = solver_settings

        floris_dict["flow_field"] = flow_field_dict
        floris_dict["farm"] = farm_dict

        # Create a new instance of floris and attach to self
        self.floris = Floris.from_dict(floris_dict)
        # Re-assign the hetergeneous inflow map to flow field
        self.floris.flow_field.het_map = self.het_map

    def get_plane_of_points(
        self,
        normal_vector="z",
        planar_coordinate=None,
    ):
        """
        Calculates velocity values through the
        :py:meth:`~.FlowField.calculate_wake` method at points in plane
        specified by inputs.

        Args:
            normal_vector (string, optional): Vector normal to plane.
                Defaults to z.
            planar_coordinate (float, optional): Value of normal vector to slice through. Defaults to None.


        Returns:
            :py:class:`pandas.DataFrame`: containing values of x1, x2, u, v, w
        """
        # Get results vectors
        x_flat = self.floris.grid.x[0, 0].flatten()
        y_flat = self.floris.grid.y[0, 0].flatten()
        z_flat = self.floris.grid.z[0, 0].flatten()
        u_flat = self.floris.flow_field.u[0, 0].flatten()
        v_flat = self.floris.flow_field.v[0, 0].flatten()
        w_flat = self.floris.flow_field.w[0, 0].flatten()

        # Create a df of these
        if normal_vector == "z":
            df = pd.DataFrame(
                {
                    "x1": x_flat,
                    "x2": y_flat,
                    "x3": z_flat,
                    "u": u_flat,
                    "v": v_flat,
                    "w": w_flat,
                }
            )
        if normal_vector == "x":
            df = pd.DataFrame(
                {
                    "x1": y_flat,
                    "x2": z_flat,
                    "x3": x_flat,
                    "u": u_flat,
                    "v": v_flat,
                    "w": w_flat,
                }
            )
        if normal_vector == "y":
            df = pd.DataFrame(
                {
                    "x1": x_flat,
                    "x2": z_flat,
                    "x3": y_flat,
                    "u": u_flat,
                    "v": v_flat,
                    "w": w_flat,
                }
            )

        # Subset to plane
        # TODO: Seems sloppy as need more than one plane in the z-direction for GCH
        if planar_coordinate is not None:
            df = df[np.isclose(df.x3, planar_coordinate)] # , atol=0.1, rtol=0.0)]

        # Drop duplicates
        # TODO is this still needed now that we setup a grid for just this plane?
        df = df.drop_duplicates()

        # Sort values of df to make sure plotting is acceptable
        df = df.sort_values(["x2", "x1"]).reset_index(drop=True)

        return df

    def calculate_horizontal_plane(
        self,
        height=None,
        x_resolution=200,
        y_resolution=200,
        x_bounds=None,
        y_bounds=None,
        wd=None,
        ws=None,
        yaw_angles=None,
    ):
        """
        Shortcut method to instantiate a :py:class:`~.tools.cut_plane.CutPlane`
        object containing the velocity field in a horizontal plane cut through
        the simulation domain at a specific height.

        Args:
            height (float): Height of cut plane. Defaults to Hub-height.
            x_resolution (float, optional): Output array resolution.
                Defaults to 200 points.
            y_resolution (float, optional): Output array resolution.
                Defaults to 200 points.
            x_bounds (tuple, optional): Limits of output array (in m).
                Defaults to None.
            y_bounds (tuple, optional): Limits of output array (in m).
                Defaults to None.

        Returns:
            :py:class:`~.tools.cut_plane.CutPlane`: containing values
            of x, y, u, v, w
        """
        #TODO update docstring
        if wd is None:
            wd = self.floris.flow_field.wind_directions
        if ws is None:
            ws = self.floris.flow_field.wind_speeds
        self.check_wind_condition_for_viz(wd=wd, ws=ws)

        # If height not provided, use the hub height
        if height is None:
            height = self.floris.turbine.hub_height
            self.logger.info("Default to hub height = %.1f for horizontal plane." % height)

        # Store the current state for reinitialization
        floris_dict = self.floris.as_dict()

        # Set the solver to a flow field planar grid
        solver_settings = {
            "type": "flow_field_planar_grid",
            "normal_vector": "z",
            "planar_coordinate": height,
            "flow_field_grid_points": [x_resolution, y_resolution],
            "flow_field_bounds": [x_bounds, y_bounds],
        }
        self.reinitialize(
            wind_directions=wd, wind_speeds=ws, solver_settings=solver_settings
        )

        # TODO this has to be done here as it seems to be lost with reinitialize
        if yaw_angles is not None:
            self.floris.farm.yaw_angles = yaw_angles

        # Calculate wake
        self.floris.solve_for_viz()

        # Get the points of data in a dataframe
        # TODO this just seems to be flattening and storing the data in a df; is this necessary? It seems the biggest depenedcy is on CutPlane and the subsequent visualization tools.
        df = self.get_plane_of_points(
            normal_vector="z",
            planar_coordinate=height,
        )

        # Compute the cutplane
        horizontal_plane = CutPlane(df, self.floris.grid.grid_resolution[0], self.floris.grid.grid_resolution[1])

        # Reset the fi object back to the turbine grid configuration
        self.floris = Floris.from_dict(floris_dict)
        self.floris.flow_field.het_map = self.het_map

        # Run the simulation again for futher postprocessing (i.e. now we can get farm power)
        self.calculate_wake()

        return horizontal_plane

    def calculate_cross_plane(
        self,
        downstream_dist=None,
        y_resolution=200,
        z_resolution=200,
        y_bounds=None,
        z_bounds=None,
        wd=None,
        ws=None,
        yaw_angles=None,
    ):
        """
        Shortcut method to instantiate a :py:class:`~.tools.cut_plane.CutPlane`
        object containing the velocity field in a horizontal plane cut through
        the simulation domain at a specific height.

        Args:
            height (float): Height of cut plane. Defaults to Hub-height.
            x_resolution (float, optional): Output array resolution.
                Defaults to 200 points.
            y_resolution (float, optional): Output array resolution.
                Defaults to 200 points.
            x_bounds (tuple, optional): Limits of output array (in m).
                Defaults to None.
            y_bounds (tuple, optional): Limits of output array (in m).
                Defaults to None.

        Returns:
            :py:class:`~.tools.cut_plane.CutPlane`: containing values
            of x, y, u, v, w
        """
        # TODO update docstring
        if wd is None:
            wd = self.floris.flow_field.wind_directions
        if ws is None:
            ws = self.floris.flow_field.wind_speeds

        self.check_wind_condition_for_viz(wd=wd, ws=ws)
        # If downstream distance is not provided, use the default
        if downstream_dist is None:
            downstream_dist = 5 * 126.0

        # Store the current state for reinitialization
        floris_dict = self.floris.as_dict()

        # Set the solver to a flow field planar grid
        solver_settings = {
            "type": "flow_field_planar_grid",
            "normal_vector": "x",
            "planar_coordinate": downstream_dist,
            "flow_field_grid_points": [y_resolution, z_resolution],
            "flow_field_bounds": [y_bounds, z_bounds],
        }
        self.reinitialize(
            wind_directions=wd, wind_speeds=ws, solver_settings=solver_settings
        )

        # TODO this has to be done here as it seems to be lost with reinitialize
        if yaw_angles is not None:
            self.floris.farm.yaw_angles = yaw_angles

        # Calculate wake
        self.floris.solve_for_viz()

        # Get the points of data in a dataframe
        # TODO this just seems to be flattening and storing the data in a df; is this necessary? It seems the biggest depenedcy is on CutPlane and the subsequent visualization tools.
        df = self.get_plane_of_points(
            normal_vector="x",
            planar_coordinate=downstream_dist,
        )

        # Compute the cutplane
        cross_plane = CutPlane(df, y_resolution, z_resolution)

        # Reset the fi object back to the turbine grid configuration
        self.floris = Floris.from_dict(floris_dict)
        self.floris.flow_field.het_map = self.het_map

        # Run the simulation again for futher postprocessing (i.e. now we can get farm power)
        self.calculate_wake()

        return cross_plane

    def calculate_y_plane(
        self,
        crossstream_dist=None,
        x_resolution=200,
        z_resolution=200,
        x_bounds=None,
        z_bounds=None,
        wd=None,
        ws=None,
        yaw_angles=None,
    ):
        """
        Shortcut method to instantiate a :py:class:`~.tools.cut_plane.CutPlane`
        object containing the velocity field in a horizontal plane cut through
        the simulation domain at a specific height.

        Args:
            height (float): Height of cut plane. Defaults to Hub-height.
            x_resolution (float, optional): Output array resolution.
                Defaults to 200 points.
            y_resolution (float, optional): Output array resolution.
                Defaults to 200 points.
            x_bounds (tuple, optional): Limits of output array (in m).
                Defaults to None.
            y_bounds (tuple, optional): Limits of output array (in m).
                Defaults to None.

        Returns:
            :py:class:`~.tools.cut_plane.CutPlane`: containing values
            of x, y, u, v, w
        """
        #TODO update docstring
        if wd is None:
            wd = self.floris.flow_field.wind_directions
        if ws is None:
            ws = self.floris.flow_field.wind_speeds
        self.check_wind_condition_for_viz(wd=wd, ws=ws)

        # If crossstream distance is not provided, use the default
        if crossstream_dist is None:
            crossstream_dist = 0.0

        # Store the current state for reinitialization
        floris_dict = self.floris.as_dict()

        # Set the solver to a flow field planar grid
        solver_settings = {
            "type": "flow_field_planar_grid",
            "normal_vector": "y",
            "planar_coordinate": crossstream_dist,
            "flow_field_grid_points": [x_resolution, z_resolution],
            "flow_field_bounds": [x_bounds, z_bounds],
        }
        self.reinitialize(
            wind_directions=wd, wind_speeds=ws, solver_settings=solver_settings
        )

        # TODO this has to be done here as it seems to be lost with reinitialize
        if yaw_angles is not None:
            self.floris.farm.yaw_angles = yaw_angles

        # Calculate wake
        self.floris.solve_for_viz()

        # Get the points of data in a dataframe
        # TODO this just seems to be flattening and storing the data in a df; is this necessary? It seems the biggest depenedcy is on CutPlane and the subsequent visualization tools.
        df = self.get_plane_of_points(
            normal_vector="y",
            planar_coordinate=crossstream_dist,
        )

        # Compute the cutplane
        y_plane = CutPlane(df, x_resolution, z_resolution)

        # Reset the fi object back to the turbine grid configuration
        self.floris = Floris.from_dict(floris_dict)
        self.floris.flow_field.het_map = self.het_map

        # Run the simulation again for futher postprocessing (i.e. now we can get farm power)
        self.calculate_wake()

        return y_plane
        
    def check_wind_condition_for_viz(self, wd=None, ws=None):
        if len(wd) > 1 or len(wd) < 1:
            raise ValueError("Wind direction input must be of length 1 for visualization. Current length is {}.".format(len(wd)))
        
        if len(ws) > 1 or len(ws) < 1:
            raise ValueError("Wind speed input must be of length 1 for visualization. Current length is {}.".format(len(ws)))

    def get_turbine_powers(self) -> NDArrayFloat:
        """Calculates the power at each turbine in the windfarm.

        Returns:
            NDArrayFloat: [description]
        """
        turbine_powers = power(
            air_density=self.floris.flow_field.air_density,
            velocities=self.floris.flow_field.u,
            yaw_angle=self.floris.farm.yaw_angles,
            pP=self.floris.turbine.pP,
            power_interp=self.floris.turbine.power_interp,
        )
        return turbine_powers

    def get_farm_power(
        self,
        include_unc=False,
        unc_pmfs=None,
        unc_options=None,
        no_wake=False,
        use_turbulence_correction=False,
    ):
        """
        Report wind plant power from instance of floris. Optionally includes
        uncertainty in wind direction and yaw position when determining power.
        Uncertainty is included by computing the mean wind farm power for a
        distribution of wind direction and yaw position deviations from the
        original wind direction and yaw angles.

        Args:
            include_unc (bool): When *True*, uncertainty in wind direction
                and/or yaw position is included when determining wind farm
                power. Defaults to *False*.
            unc_pmfs (dictionary, optional): A dictionary containing optional
                probability mass functions describing the distribution of wind
                direction and yaw position deviations when wind direction and/or
                yaw position uncertainty is included in the power calculations.
                Contains the following key-value pairs:

                -   **wd_unc** (*np.array*): Wind direction deviations from the
                    original wind direction.
                -   **wd_unc_pmf** (*np.array*): Probability of each wind
                    direction deviation in **wd_unc** occuring.
                -   **yaw_unc** (*np.array*): Yaw angle deviations from the
                    original yaw angles.
                -   **yaw_unc_pmf** (*np.array*): Probability of each yaw angle
                    deviation in **yaw_unc** occuring.

                Defaults to None, in which case default PMFs are calculated
                using values provided in **unc_options**.
            unc_options (dictionary, optional): A dictionary containing values
                used to create normally-distributed, zero-mean probability mass
                functions describing the distribution of wind direction and yaw
                position deviations when wind direction and/or yaw position
                uncertainty is included. This argument is only used when
                **unc_pmfs** is None and contains the following key-value pairs:

                -   **std_wd** (*float*): A float containing the standard
                    deviation of the wind direction deviations from the
                    original wind direction.
                -   **std_yaw** (*float*): A float containing the standard
                    deviation of the yaw angle deviations from the original yaw
                    angles.
                -   **pmf_res** (*float*): A float containing the resolution in
                    degrees of the wind direction and yaw angle PMFs.
                -   **pdf_cutoff** (*float*): A float containing the cumulative
                    distribution function value at which the tails of the
                    PMFs are truncated.

                Defaults to None. Initializes to {'std_wd': 4.95, 'std_yaw':
                1.75, 'pmf_res': 1.0, 'pdf_cutoff': 0.995}.
            no_wake: (bool, optional): When *True* updates the turbine
                quantities without calculating the wake or adding the
                wake to the flow field. Defaults to *False*.
            use_turbulence_correction: (bool, optional): When *True* uses a
                turbulence parameter to adjust power output calculations.
                Defaults to *False*.

        Returns:
            float: Sum of wind turbine powers.
        """
        # TODO: Turbulence correction used in the power calculation, but may not be in
        # the model yet
        # TODO: Turbines need a switch for using turbulence correction
        # TODO: Uncomment out the following two lines once the above are resolved
        # for turbine in self.floris.farm.turbines:
        #     turbine.use_turbulence_correction = use_turbulence_correction

        if include_unc:
            unc_pmfs = _generate_uncertainty_parameters(unc_options, unc_pmfs)

            # TODO: The original form of this is:
            # self.floris.farm.wind_map.input_direction[0], but it's unclear why we're
            # capping at just the first wind direction. Should this behavior be kept?
            # I'm unsure as to how the first wind direction is the original, so it could
            # just be a naming thing that's throwing me off....
            wd_orig = self.floris.flow_field.wind_directions

            yaw_angles = self.floris.farm.yaw_angles()
            self.reinitialize(wind_direction=wd_orig + unc_pmfs["wd_unc"])
            power_at_yaw = [
                self.get_farm_power_for_yaw_angle(yaw_angles + delta_yaw, no_wake=no_wake)
                for delta_yaw in unc_pmfs["yaw_unc"]
            ]
            mean_farm_power = unc_pmfs["wd_unc_pmf"] * unc_pmfs["yaw_unc_pmf"] * np.array(power_at_yaw)

            # reinitialize with original values
            self.reinitialize(wind_direction=wd_orig)
            self.calculate_wake(yaw_angles=yaw_angles, no_wake=no_wake)
            return mean_farm_power

        turbine_powers = self.get_turbine_powers()
        # for i in range(self.floris.farm.n_turbines):
        #     avg_v_i = average_velocity(
        #         self.floris.flow_field.u,
        #         ix_filter=[i]
        #     )
        #     power_i = power(
        #         air_density=self.floris.flow_field.air_density,
        #         velocities=self.floris.flow_field.u,
        #         yaw_angle=self.floris.farm.yaw_angles,
        #         pP=self.floris.turbine.pP,
        #         power_interp=self.floris.turbine.power_interp,
        #         ix_filter=[i]
        #     )
        #     print(i, avg_v_i[0,0,0], power_i[0,0,0])
        # print( "total, MW", np.sum(turbine_powers[0,0]) / 1000000 )
        return np.sum(turbine_powers, axis=2)

    def get_farm_AEP(
        self,
        wd: NDArrayFloat | list[float],
        ws: NDArrayFloat | list[float],
        freq: NDArrayFloat | list[list[float]],
        yaw: NDArrayFloat | list[float] | None = None,
        limit_ws: bool = False,
        ws_limit_tol: float = 0.001,
        ws_cutout: float = 30.0,
    ) -> float:
        """
        Estimate annual energy production (AEP) for distributions of wind speed, wind
        direction, wind rose probability, and yaw offset. This can be computed for
        pre-determined wind direction and wind speed combinations, as was the case in
        FLORIS v2, or additionally, the unique wind directions, wind speeds, and their
        probabilities can be input.

        Args:
            wd (NDArrayFloat | list[float]): List or array of wind direction values.
                Either a unique list of wind directions can be used or the wind
                directions corresponding to a pre-computed set of combinations
                should be used.
            ws (NDArrayFloat | list[float]): List or array of wind speed values.
                Either a unique list of wind speeds can be used or the wind speeds
                corresponding to a pre-computed set of combinations should be used.
            freq (NDArrayFloat | list[list[float]]): Frequencies corresponding to either
                the pre-computed combinations of wind directions and wind speeds or the
                full wind rose with dimensions (N wind directions x N wind speeds).
            yaw (NDArrayFloat | list[float] | None, optional): List or array of yaw
                values if wake is steering implemented that correspond with the number
                of wind directions. Defaults to None.
            limit_ws (bool, optional): When *True*, detect wind speed when power
                reaches it's maximum value for a given wind direction. For all
                higher wind speeds, use last calculated value when below cut
                out. Defaults to False.
            ws_limit_tol (float, optional): Tolerance fraction for determining
                wind speed where power stops changing. If limit_ws is *True*,
                assume power remains constant up to cut out for wind speeds
                above the point where power changes less than ws_limit_tol of
                the previous power. Defaults to 0.001.
            ws_cutout (float, optional): Cut out wind speed (m/s). If limit_ws
                is *True*, assume power is zero for wind speeds greater than or
                equal to ws_cutout. Defaults to 30.0

        Returns:
            float: AEP for wind farm.
        """

        # # Convert the required inputs to arrays
        # wd = np.array(wd)
        # ws = np.array(ws)
        # freq = np.array(freq)

        # # Determine if the direction and speed inputs provided are a set of pre-determined
        # # combinations, and compute the full combination set if so, where the value
        # # in freq will be set to 0 if a combination was not in the original set to ensure
        # # it's not counted.
        # wd_unique = wd.unique()
        # ws_unique = ws.unique()
        # if np.array_equal(wd_unique, sorted(wd)) and np.array_equal(ws_unique, sorted(ws)):
        #     # Reshape the frequency input if required, and leave the unique inputs as-is
        #     if freq.shape != (wd_unique.size, ws_unique.size):
        #         freq = freq.reshape((wd_unique.size, ws_unique.size))
        # else:
        #     # Compute all the combinations
        #     wd_unique, ws_unique, freq, yaw = correct_for_all_combinations(wd, ws, freq, yaw)

        # # If the yaw input is still None, then create a None array as inputs
        # if yaw is None:
        #     N = wd_unique.size * ws_unique.size
        #     yaw = np.array([None] * N).reshape(wd_unique.size, ws_unique.size)
        # else:
        #     yaw = np.array(yaw)

        # filter out wind speeds beyond the cutoff, if necessary
        # if limit_ws:
        #     ix_ws_filter = ws_unique >= ws_cutout
        #     ws_unique = ws_unique[ix_ws_filter]
        #     freq = freq[:, ix_ws_filter]
        #     yaw = yaw[:, ix_ws_filter]

        # self.reinitialize(wind_direction=wd_unique, wind_speed=ws_unique, wind_rose_probability=freq)
        self.calculate_wake()
        farm_power = self.get_farm_power()  # TODO: Do we need to specify an axis since this is a sum?
        AEP = farm_power * freq * 8760
        return np.sum(AEP)

    def _calc_one_AEP_case(self, wd, ws, freq, yaw=None):
        self.reinitialize(wind_direction=[wd], wind_speed=[ws])
        self.calculate_wake(yaw_angles=yaw)
        return self.get_farm_power() * freq * 8760

    # def get_farm_AEP_parallel(
    #     self,
    #     wd: NDArrayFloat | list[float],
    #     ws: NDArrayFloat | list[float],
    #     freq: NDArrayFloat | list[list[float]],
    #     yaw: NDArrayFloat | list[float] | None = None,
    #     jobs=-1,
    # ):
    #     """
    #     Estimate annual energy production (AEP) for distributions of wind
    #     speed, wind direction and yaw offset with parallel computations on
    #     a single comptuer.

    #     # TODO: Update the docstrings and allow for the use of precomputed combinations
    #     as well as unique inputs that need to be computed. Same for the other AEPs

    #     Args:
    #         wd (iterable): List or array of wind direction values.
    #         ws (iterable): List or array of wind speed values.
    #         freq (iterable): Frequencies corresponding to wind direction and wind speed
    #             combinations in the wind rose with, shape (N wind directions x N wind speeds).
    #         yaw (iterable, optional): List or array of yaw values if wake is steering
    #             implemented, with shape (N wind directions). Defaults to None.
    #         jobs (int, optional): The number of jobs (cores) to use in the parallel
    #             computations.

    #     Returns:
    #         float: AEP for wind farm.
    #     """
    #     if jobs < -1:
    #         raise ValueError("Input 'jobs' cannot be negative.")
    #     if jobs == -1:
    #         jobs = int(np.ceil(cpu_count() * 0.8))
    #     if jobs > 0:
    #         jobs = min(jobs, cpu_count())
    #     if jobs > len(wd):
    #         jobs = len(wd)

    #     if yaw is None:
    #         yaw = [None] * len(wd)

    #     wd = np.array(wd)
    #     ws = np.array(ws)
    #     freq = np.array(freq)

    #     # Make one large list of arguments, then flatten and resort the nested tuples
    #     # to the correct ordering of self, wd, ws, freq, yaw
    #     global_arguments = list(zip(repeat(self), zip(wd, yaw), ws, freq.flatten()))
    #     # OR is this supposed to be all wind speeds for each wind direction?:
    #     # global_arguments = list(zip(repeat(self), zip(wd, yaw), repeat(ws), freq))
    #     # global_arguments = [(s, n[0], wspd, f, n[1]) for s, n, wspd, f in global_arguments]
    #     global_arguments = [(s, n[0][0], n[1], f, n[0][1]) for s, n, f in global_arguments]

    #     num_cases = wd.size * ws.size
    #     chunksize = int(np.ceil(num_cases / jobs))

    #     with Pool(jobs) as pool:
    #         opt = pool.starmap(global_calc_one_AEP_case, global_arguments, chunksize=chunksize)
    #         # add AEP to overall AEP

    #     return 0.0 + np.sum(opt)

    def calculate_AEP_wind_limit(self, num_turbines, x_spacing, start_ws, threshold):
        orig_layout_x = self.layout_x
        orig_layout_y = self.layout_y
        D = self.floris.farm.turbines[0].rotor_diameter

        self.reinitialize(
            layout_array=(
                [i * x_spacing * D for i in range(num_turbines)],
                [0.0] * num_turbines,
            ),
            wind_speed=start_ws,
        )
        self.calculate_wake()

        prev_power = 1.0
        cur_power = self.get_farm_power()
        ws = start_ws

        while np.abs(prev_power - cur_power) / prev_power > threshold:
            prev_power = cur_power
            ws += 0.2
            self.reinitialize(wind_speed=ws)
            self.calculate_wake()
            cur_power = self.get_farm_power()
        ws += 1.0

        self.reinitialize(layout_array=(orig_layout_x, orig_layout_y), wind_speed=ws)
        self.calculate_wake()
        self.max_power = self.get_farm_power()
        self.ws_limit = ws

    @property
    def layout_x(self):
        """
        Wind turbine coordinate information.

        Returns:
            np.array: Wind turbine x-coordinate.
        """
        return self.floris.farm.layout_x

    @property
    def layout_y(self):
        """
        Wind turbine coordinate information.

        Returns:
            np.array: Wind turbine y-coordinate.
        """
        return self.floris.farm.layout_y


    def get_turbine_layout(self, z=False):
        """
        Get turbine layout

        Args:
            z (bool): When *True*, return lists of x, y, and z coords,
            otherwise, return x and y only. Defaults to *False*.

        Returns:
            np.array: lists of x, y, and (optionally) z coordinates of
                      each turbine
        """
        xcoords, ycoords, zcoords = np.array([c.elements for c in self.floris.farm.coordinates]).T
        if z:
            return xcoords, ycoords, zcoords
        else:
            return xcoords, ycoords



def generate_heterogeneous_wind_map(speed_ups, x, y, z=None):
    if z is not None:
        # Compute the 3-dimensional interpolants for each wind diretion
        # Linear interpolation is used for points within the user-defined area of values,
        # while a nearest-neighbor interpolant is used for points outside that region
        in_region = [LinearNDInterpolator(list(zip(x, y, z)), speed_up, fill_value=np.nan) for speed_up in speed_ups]
        out_region = [NearestNDInterpolator(list(zip(x, y, z)), speed_up) for speed_up in speed_ups]
    else:
        # Compute the 2-dimensional interpolants for each wind diretion
        # Linear interpolation is used for points within the user-defined area of values,
        # while a nearest-neighbor interpolant is used for points outside that region
        in_region = [LinearNDInterpolator(list(zip(x, y)), speed_up, fill_value=np.nan) for speed_up in speed_ups]
        out_region = [NearestNDInterpolator(list(zip(x, y)), speed_up) for speed_up in speed_ups]

    return [in_region, out_region]

# def global_calc_one_AEP_case(FlorisInterface, wd, ws, freq, yaw=None):
#     return FlorisInterface._calc_one_AEP_case(wd, ws, freq, yaw)

DEFAULT_UNCERTAINTY = {"std_wd": 4.95, "std_yaw": 1.75, "pmf_res": 1.0, "pdf_cutoff": 0.995}


def _generate_uncertainty_parameters(unc_options: dict, unc_pmfs: dict) -> dict:
    """Generates the uncertainty parameters for `FlorisInterface.get_farm_power` and
    `FlorisInterface.get_turbine_power` for more details.

    Args:
        unc_options (dict): See `FlorisInterface.get_farm_power` or `FlorisInterface.get_turbine_power`.
        unc_pmfs (dict): See `FlorisInterface.get_farm_power` or `FlorisInterface.get_turbine_power`.

    Returns:
        dict: [description]
    """
    if (unc_options is None) & (unc_pmfs is None):
        unc_options = DEFAULT_UNCERTAINTY

    if unc_pmfs is not None:
        return unc_pmfs

    wd_unc = np.zeros(1)
    wd_unc_pmf = np.ones(1)
    yaw_unc = np.zeros(1)
    yaw_unc_pmf = np.ones(1)

    # create normally distributed wd and yaw uncertaitny pmfs if appropriate
    if unc_options["std_wd"] > 0:
        wd_bnd = int(np.ceil(norm.ppf(unc_options["pdf_cutoff"], scale=unc_options["std_wd"]) / unc_options["pmf_res"]))
        bound = wd_bnd * unc_options["pmf_res"]
        wd_unc = np.linspace(-1 * bound, bound, 2 * wd_bnd + 1)
        wd_unc_pmf = norm.pdf(wd_unc, scale=unc_options["std_wd"])
        wd_unc_pmf /= np.sum(wd_unc_pmf)  # normalize so sum = 1.0

    if unc_options["std_yaw"] > 0:
        yaw_bnd = int(
            np.ceil(norm.ppf(unc_options["pdf_cutoff"], scale=unc_options["std_yaw"]) / unc_options["pmf_res"])
        )
        bound = yaw_bnd * unc_options["pmf_res"]
        yaw_unc = np.linspace(-1 * bound, bound, 2 * yaw_bnd + 1)
        yaw_unc_pmf = norm.pdf(yaw_unc, scale=unc_options["std_yaw"])
        yaw_unc_pmf /= np.sum(yaw_unc_pmf)  # normalize so sum = 1.0

    unc_pmfs = {
        "wd_unc": wd_unc,
        "wd_unc_pmf": wd_unc_pmf,
        "yaw_unc": yaw_unc,
        "yaw_unc_pmf": yaw_unc_pmf,
    }
    return unc_pmfs


# def correct_for_all_combinations(
#     wd: NDArrayFloat,
#     ws: NDArrayFloat,
#     freq: NDArrayFloat,
#     yaw: NDArrayFloat | None = None,
# ) -> tuple[NDArrayFloat]:
#     """Computes the probabilities for the complete windrose from the desired wind
#     direction and wind speed combinations and their associated probabilities so that
#     any undesired combinations are filled with a 0.0 probability.

#     Args:
#         wd (NDArrayFloat): List or array of wind direction values.
#         ws (NDArrayFloat): List or array of wind speed values.
#         freq (NDArrayFloat): Frequencies corresponding to wind
#             speeds and directions in wind rose with dimensions
#             (N wind directions x N wind speeds).
#         yaw (NDArrayFloat | None): The corresponding yaw angles for each of the wind
#             direction and wind speed combinations, or None. Defaults to None.

#     Returns:
#         NDArrayFloat, NDArrayFloat, NDArrayFloat: The unique wind directions, wind
#             speeds, and the associated probability of their combination combinations in
#             an array of shape (N wind directions x N wind speeds).
#     """

#     combos_to_compute = np.array(list(zip(wd, ws, freq)))

#     unique_wd = wd.unique()
#     unique_ws = ws.unique()
#     all_combos = np.array(list(product(unique_wd, unique_ws)), dtype=float)
#     all_combos = np.hstack((all_combos, np.zeros((all_combos.shape[0], 1), dtype=float)))
#     expanded_yaw = np.array([None] * all_combos.shape[0]).reshape(unique_wd.size, unique_ws.size)

#     ix_match = [np.where((all_combos[:, :2] == combo[:2]).all(1))[0][0] for combo in combos_to_compute]
#     all_combos[ix_match, 2] = combos_to_compute[:, 2]
#     if yaw is not None:
#         expanded_yaw[ix_match] = yaw
#     freq = all_combos.T[2].reshape((unique_wd.size, unique_ws.size))
#     return unique_wd, unique_ws, freq


    # def get_set_of_points(self, x_points, y_points, z_points):
    #     """
    #     Calculates velocity values through the
    #     :py:meth:`~.FlowField.calculate_wake` method at points specified by
    #     inputs.

    #     Args:
    #         x_points (float): X-locations to get velocity values at.
    #         y_points (float): Y-locations to get velocity values at.
    #         z_points (float): Z-locations to get velocity values at.

    #     Returns:
    #         :py:class:`pandas.DataFrame`: containing values of x, y, z, u, v, w
    #     """
    #     # Get a copy for the flow field so don't change underlying grid points
    #     flow_field = copy.deepcopy(self.floris.flow_field)

    #     if hasattr(self.floris.wake.velocity_model, "requires_resolution"):
    #         if self.floris.velocity_model.requires_resolution:

    #             # If this is a gridded model, must extract from full flow field
    #             self.logger.info(
    #                 "Model identified as %s requires use of underlying grid print"
    #                 % self.floris.wake.velocity_model.model_string
    #             )
    #             self.logger.warning("FUNCTION NOT AVAILABLE CURRENTLY")

    #     # Set up points matrix
    #     points = np.row_stack((x_points, y_points, z_points))

    #     # TODO: Calculate wake inputs need to be mapped
    #     raise_error = True
    #     if raise_error:
    #         raise NotImplementedError("Additional point calculation is not yet supported!")
    #     # Recalculate wake with these points
    #     flow_field.calculate_wake(points=points)

    #     # Get results vectors
    #     x_flat = flow_field.x.flatten()
    #     y_flat = flow_field.y.flatten()
    #     z_flat = flow_field.z.flatten()
    #     u_flat = flow_field.u.flatten()
    #     v_flat = flow_field.v.flatten()
    #     w_flat = flow_field.w.flatten()

    #     df = pd.DataFrame(
    #         {
    #             "x": x_flat,
    #             "y": y_flat,
    #             "z": z_flat,
    #             "u": u_flat,
    #             "v": v_flat,
    #             "w": w_flat,
    #         }
    #     )

    #     # Subset to points requests
    #     df = df[df.x.isin(x_points)]
    #     df = df[df.y.isin(y_points)]
    #     df = df[df.z.isin(z_points)]

    #     # Drop duplicates
    #     df = df.drop_duplicates()

    #     # Return the dataframe
    #     return df
    
    # def get_flow_data(self, resolution=None, grid_spacing=10, velocity_deficit=False):
    #     """
    #     Generate :py:class:`~.tools.flow_data.FlowData` object corresponding to
    #     active FLORIS instance.

    #     Velocity and wake models requiring calculation on a grid implement a
    #     discretized domain at resolution **grid_spacing**. This is distinct
    #     from the resolution of the returned flow field domain.

    #     Args:
    #         resolution (float, optional): Resolution of output data.
    #             Only used for wake models that require spatial
    #             resolution (e.g. curl). Defaults to None.
    #         grid_spacing (int, optional): Resolution of grid used for
    #             simulation. Model results may be sensitive to resolution.
    #             Defaults to 10.
    #         velocity_deficit (bool, optional): When *True*, normalizes velocity
    #             with respect to initial flow field velocity to show relative
    #             velocity deficit (%). Defaults to *False*.

    #     Returns:
    #         :py:class:`~.tools.flow_data.FlowData`: FlowData object
    #     """

    #     if resolution is None:
    #         if not self.floris.wake.velocity_model.requires_resolution:
    #             self.logger.info("Assuming grid with spacing %d" % grid_spacing)
    #             (
    #                 xmin,
    #                 xmax,
    #                 ymin,
    #                 ymax,
    #                 zmin,
    #                 zmax,
    #             ) = self.floris.flow_field.domain_bounds  # TODO: No grid attribute within FlowField
    #             resolution = Vec3(
    #                 1 + (xmax - xmin) / grid_spacing,
    #                 1 + (ymax - ymin) / grid_spacing,
    #                 1 + (zmax - zmin) / grid_spacing,
    #             )
    #         else:
    #             self.logger.info("Assuming model resolution")
    #             resolution = self.floris.wake.velocity_model.model_grid_resolution

    #     # Get a copy for the flow field so don't change underlying grid points
    #     flow_field = copy.deepcopy(self.floris.flow_field)

    #     if (
    #         flow_field.wake.velocity_model.requires_resolution
    #         and flow_field.wake.velocity_model.model_grid_resolution != resolution
    #     ):
    #         self.logger.warning(
    #             "WARNING: The current wake velocity model contains a "
    #             + "required grid resolution; the Resolution given to "
    #             + "FlorisInterface.get_flow_field is ignored."
    #         )
    #         resolution = flow_field.wake.velocity_model.model_grid_resolution
    #     flow_field.reinitialize(with_resolution=resolution)  # TODO: Not implemented
    #     self.logger.info(resolution)
    #     # print(resolution)
    #     flow_field.steady_state_atmospheric_condition()

    #     order = "f"
    #     x = flow_field.x.flatten(order=order)
    #     y = flow_field.y.flatten(order=order)
    #     z = flow_field.z.flatten(order=order)

    #     u = flow_field.u.flatten(order=order)
    #     v = flow_field.v.flatten(order=order)
    #     w = flow_field.w.flatten(order=order)

    #     # find percent velocity deficit
    #     if velocity_deficit:
    #         u = abs(u - flow_field.u_initial.flatten(order=order)) / flow_field.u_initial.flatten(order=order) * 100
    #         v = abs(v - flow_field.v_initial.flatten(order=order)) / flow_field.v_initial.flatten(order=order) * 100
    #         w = abs(w - flow_field.w_initial.flatten(order=order)) / flow_field.w_initial.flatten(order=order) * 100

    #     # Determine spacing, dimensions and origin
    #     unique_x = np.sort(np.unique(x))
    #     unique_y = np.sort(np.unique(y))
    #     unique_z = np.sort(np.unique(z))
    #     spacing = Vec3(
    #         unique_x[1] - unique_x[0],
    #         unique_y[1] - unique_y[0],
    #         unique_z[1] - unique_z[0],
    #     )
    #     dimensions = Vec3(len(unique_x), len(unique_y), len(unique_z))
    #     origin = Vec3(0.0, 0.0, 0.0)
    #     return FlowData(x, y, z, u, v, w, spacing=spacing, dimensions=dimensions, origin=origin)


    # def get_turbine_power(
    #     self,
    #     include_unc=False,
    #     unc_pmfs=None,
    #     unc_options=None,
    #     no_wake=False,
    #     use_turbulence_correction=False,
    # ):
    #     """
    #     Report power from each wind turbine.

    #     Args:
    #         include_unc (bool): If *True*, uncertainty in wind direction
    #             and/or yaw position is included when determining turbine
    #             powers. Defaults to *False*.
    #         unc_pmfs (dictionary, optional): A dictionary containing optional
    #             probability mass functions describing the distribution of wind
    #             direction and yaw position deviations when wind direction and/or
    #             yaw position uncertainty is included in the power calculations.
    #             Contains the following key-value pairs:

    #             -   **wd_unc** (*np.array*): Wind direction deviations from the
    #                 original wind direction.
    #             -   **wd_unc_pmf** (*np.array*): Probability of each wind
    #                 direction deviation in **wd_unc** occuring.
    #             -   **yaw_unc** (*np.array*): Yaw angle deviations from the
    #                 original yaw angles.
    #             -   **yaw_unc_pmf** (*np.array*): Probability of each yaw angle
    #                 deviation in **yaw_unc** occuring.

    #             Defaults to None, in which case default PMFs are calculated
    #             using values provided in **unc_options**.
    #         unc_options (dictionary, optional): A dictionary containing values
    #             used to create normally-distributed, zero-mean probability mass
    #             functions describing the distribution of wind direction and yaw
    #             position deviations when wind direction and/or yaw position
    #             uncertainty is included. This argument is only used when
    #             **unc_pmfs** is None and contains the following key-value pairs:

    #             -   **std_wd** (*float*): A float containing the standard
    #                 deviation of the wind direction deviations from the
    #                 original wind direction.
    #             -   **std_yaw** (*float*): A float containing the standard
    #                 deviation of the yaw angle deviations from the original yaw
    #                 angles.
    #             -   **pmf_res** (*float*): A float containing the resolution in
    #                 degrees of the wind direction and yaw angle PMFs.
    #             -   **pdf_cutoff** (*float*): A float containing the cumulative
    #                 distribution function value at which the tails of the
    #                 PMFs are truncated.

    #             Defaults to None. Initializes to {'std_wd': 4.95, 'std_yaw': 1.
    #             75, 'pmf_res': 1.0, 'pdf_cutoff': 0.995}.
    #         no_wake: (bool, optional): When *True* updates the turbine
    #             quantities without calculating the wake or adding the
    #             wake to the flow field. Defaults to *False*.
    #         use_turbulence_correction: (bool, optional): When *True* uses a
    #             turbulence parameter to adjust power output calculations.
    #             Defaults to *False*.

    #     Returns:
    #         np.array: Power produced by each wind turbine.
    #     """
    #     # TODO: Turbulence correction used in the power calculation, but may not be in
    #     # the model yet
    #     # TODO: Turbines need a switch for using turbulence correction
    #     # TODO: Uncomment out the following two lines once the above are resolved
    #     # for turbine in self.floris.farm.turbines:
    #     #     turbine.use_turbulence_correction = use_turbulence_correction

    #     if include_unc:
    #         unc_pmfs = _generate_uncertainty_parameters(unc_options, unc_pmfs)

    #         mean_farm_power = np.zeros(self.floris.farm.n_turbines)
    #         wd_orig = self.floris.flow_field.wind_directions  # TODO: same comment as in get_farm_power

    #         yaw_angles = self.get_yaw_angles()
    #         self.reinitialize(wind_direction=wd_orig[0] + unc_pmfs["wd_unc"])
    #         for i, delta_yaw in enumerate(unc_pmfs["yaw_unc"]):
    #             self.calculate_wake(
    #                 yaw_angles=list(np.array(yaw_angles) + delta_yaw),
    #                 no_wake=no_wake,
    #             )
    #             mean_farm_power += unc_pmfs["wd_unc_pmf"] * unc_pmfs["yaw_unc_pmf"][i] * self._get_turbine_powers()

    #         # reinitialize with original values
    #         self.reinitialize(wind_direction=wd_orig)
    #         self.calculate_wake(yaw_angles=yaw_angles, no_wake=no_wake)
    #         return mean_farm_power

    #     return self._get_turbine_powers()

    # def get_power_curve(self, wind_speeds):
    #     """
    #     Return the power curve given a set of wind speeds

    #     Args:
    #         wind_speeds (np.array): array of wind speeds to get power curve
    #     """

    #     # TODO: Why is this done? Should we expand for evenutal multiple turbines types
    #     # or just allow a filter on the turbine index?
    #     # Temporarily set the farm to a single turbine
    #     saved_layout_x = self.layout_x
    #     saved_layout_y = self.layout_y

    #     self.reinitialize(wind_speed=wind_speeds, layout_array=([0], [0]))
    #     self.calculate_wake()
    #     turbine_power = self._get_turbine_powers()

    #     # Set it back
    #     self.reinitialize(layout_array=(saved_layout_x, saved_layout_y))

    #     return turbine_power

    # def get_farm_power_for_yaw_angle(
    #     self,
    #     yaw_angles,
    #     include_unc=False,
    #     unc_pmfs=None,
    #     unc_options=None,
    #     no_wake=False,
    # ):
    #     """
    #     Assign yaw angles to turbines, calculate wake, and report farm power.

    #     Args:
    #         yaw_angles (np.array): Yaw to apply to each turbine.
    #         include_unc (bool, optional): When *True*, includes wind direction
    #             uncertainty in estimate of wind farm power. Defaults to *False*.
    #         unc_pmfs (dictionary, optional): A dictionary containing optional
    #             probability mass functions describing the distribution of wind
    #             direction and yaw position deviations when wind direction and/or
    #             yaw position uncertainty is included in the power calculations.
    #             Contains the following key-value pairs:

    #             -   **wd_unc** (*np.array*): Wind direction deviations from the
    #                 original wind direction.
    #             -   **wd_unc_pmf** (*np.array*): Probability of each wind
    #                 direction deviation in **wd_unc** occuring.
    #             -   **yaw_unc** (*np.array*): Yaw angle deviations from the
    #                 original yaw angles.
    #             -   **yaw_unc_pmf** (*np.array*): Probability of each yaw angle
    #                 deviation in **yaw_unc** occuring.

    #             Defaults to None, in which case default PMFs are calculated
    #             using values provided in **unc_options**.
    #         unc_options (dictionary, optional): A dictionary containing values
    #             used to create normally-distributed, zero-mean probability mass
    #             functions describing the distribution of wind direction and yaw
    #             position deviations when wind direction and/or yaw position
    #             uncertainty is included. This argument is only used when
    #             **unc_pmfs** is None and contains the following key-value pairs:

    #             -   **std_wd** (*float*): A float containing the standard
    #                 deviation of the wind direction deviations from the
    #                 original wind direction.
    #             -   **std_yaw** (*float*): A float containing the standard
    #                 deviation of the yaw angle deviations from the original yaw
    #                 angles.
    #             -   **pmf_res** (*float*): A float containing the resolution in
    #                 degrees of the wind direction and yaw angle PMFs.
    #             -   **pdf_cutoff** (*float*): A float containing the cumulative
    #                 distribution function value at which the tails of the
    #                 PMFs are truncated.

    #             Defaults to None. Initializes to {'std_wd': 4.95, 'std_yaw': 1.
    #             75, 'pmf_res': 1.0, 'pdf_cutoff': 0.995}.
    #         no_wake: (bool, optional): When *True* updates the turbine
    #             quantities without calculating the wake or adding the
    #             wake to the flow field. Defaults to *False*.

    #     Returns:
    #         float: Wind plant power. #TODO negative? in kW?
    #     """

    #     self.calculate_wake(yaw_angles=yaw_angles, no_wake=no_wake)

    #     return self.get_farm_power(include_unc=include_unc, unc_pmfs=unc_pmfs, unc_options=unc_options)

    # def copy_and_update_turbine_map(
    #     self, base_turbine_id: str, update_parameters: dict, new_id: str | None = None
    # ) -> dict:
    #     """Creates a new copy of an existing turbine and updates the parameters based on
    #     user input. This function is a helper to make the v2 -> v3 transition easier.

    #     Args:
    #         base_turbine_id (str): The base turbine's ID in `floris.farm.turbine_id`.
    #         update_parameters (dict): A dictionary of the turbine parameters to update
    #             and their new valies.
    #         new_id (str, optional): The new `turbine_id`, if `None` a unique
    #             identifier will be appended to the end. Defaults to None.

    #     Returns:
    #         dict: A turbine mapping that can be passed directly to `change_turbine`.
    #     """
    #     if new_id is None:
    #         new_id = f"{base_turbine_id}_copy{self.unique_copy_id}"
    #         self.unique_copy_id += 1

    #     turbine = {new_id: self.floris.turbine[base_turbine_id]._asdict()}
    #     turbine[new_id].update(update_parameters)
    #     return turbine

    # def change_turbine(
    #     self,
    #     turbine_indices: list[int],
    #     new_turbine_map: dict[str, dict[str, Any]],
    #     update_specified_wind_height: bool = False,
    # ):
    #     """
    #     Change turbine properties for specified turbines.

    #     Args:
    #         turbine_indices (list[int]): List of turbine indices to change.
    #         new_turbine_map (dict[str, dict[str, Any]]): New dictionary of turbine
    #             parameters to create the new turbines for each of `turbine_indices`.
    #         update_specified_wind_height (bool, optional): When *True*, update specified
    #             wind height to match new hub_height. Defaults to *False*.
    #     """
    #     new_turbine = True
    #     new_turbine_id = [*new_turbine_map][0]
    #     if new_turbine_id in self.floris.farm.turbine_map:
    #         new_turbine = False
    #         self.logger.info(f"Turbines {turbine_indices} will be re-mapped to the definition for: {new_turbine_id}")

    #     self.floris.farm.turbine_id = [
    #         new_turbine_id if i in turbine_indices else t_id for i, t_id in enumerate(self.floris.farm.turbine_id)
    #     ]
    #     if new_turbine:
    #         self.logger.info(f"Turbines {turbine_indices} have been mapped to the new definition for: {new_turbine_id}")

    #     # Update the turbine mapping if a new turbine was provided, then regenerate the
    #     # farm arrays for the turbine farm
    #     if new_turbine:
    #         turbine_map = self.floris.farm._asdict()["turbine_map"]
    #         turbine_map.update(new_turbine_map)
    #         self.floris.farm.turbine_map = turbine_map
    #     self.floris.farm.generate_farm_points()

    #     new_hub_height = new_turbine_map[new_turbine_id]["hub_height"]
    #     changed_hub_height = new_hub_height != self.floris.flow_field.reference_wind_height

    #     # Alert user if changing hub-height and not specified wind height
    #     if changed_hub_height and not update_specified_wind_height:
    #         self.logger.info("Note, updating hub height but not updating " + "the specfied_wind_height")

    #     if changed_hub_height and update_specified_wind_height:
    #         self.logger.info(f"Note, specfied_wind_height changed to hub-height: {new_hub_height}")
    #         self.reinitialize(specified_wind_height=new_hub_height)

    #     # Finish by re-initalizing the flow field
    #     self.reinitialize()

    # def set_use_points_on_perimeter(self, use_points_on_perimeter=False):
    #     """
    #     Set whether to use the points on the rotor diameter (perimeter) when
    #     calculating flow field and wake.

    #     Args:
    #         use_points_on_perimeter (bool): When *True*, use points at rotor
    #             perimeter in wake and flow calculations. Defaults to *False*.
    #     """
    #     for turbine in self.floris.farm.turbines:
    #         turbine.use_points_on_perimeter = use_points_on_perimeter
    #         turbine.initialize_turbine()

    # def set_gch(self, enable=True):
    #     """
    #     Enable or disable Gauss-Curl Hybrid (GCH) functions
    #     :py:meth:`~.GaussianModel.calculate_VW`,
    #     :py:meth:`~.GaussianModel.yaw_added_recovery_correction`, and
    #     :py:attr:`~.VelocityDeflection.use_secondary_steering`.

    #     Args:
    #         enable (bool, optional): Flag whether or not to implement flow
    #             corrections from GCH model. Defaults to *True*.
    #     """
    #     self.set_gch_yaw_added_recovery(enable)
    #     self.set_gch_secondary_steering(enable)

    # def set_gch_yaw_added_recovery(self, enable=True):
    #     """
    #     Enable or Disable yaw-added recovery (YAR) from the Gauss-Curl Hybrid
    #     (GCH) model and the control state of
    #     :py:meth:`~.GaussianModel.calculate_VW_velocities` and
    #     :py:meth:`~.GaussianModel.yaw_added_recovery_correction`.

    #     Args:
    #         enable (bool, optional): Flag whether or not to implement yaw-added
    #             recovery from GCH model. Defaults to *True*.
    #     """
    #     model_params = self.get_model_parameters()
    #     use_secondary_steering = model_params["Wake Deflection Parameters"]["use_secondary_steering"]

    #     if enable:
    #         model_params["Wake Velocity Parameters"]["use_yaw_added_recovery"] = True

    #         # If enabling be sure calc vw is on
    #         model_params["Wake Velocity Parameters"]["calculate_VW_velocities"] = True

    #     if not enable:
    #         model_params["Wake Velocity Parameters"]["use_yaw_added_recovery"] = False

    #         # If secondary steering is also off, disable calculate_VW_velocities
    #         if not use_secondary_steering:
    #             model_params["Wake Velocity Parameters"]["calculate_VW_velocities"] = False

    #     self.set_model_parameters(model_params)
    #     self.reinitialize()

    # def set_gch_secondary_steering(self, enable=True):
    #     """
    #     Enable or Disable secondary steering (SS) from the Gauss-Curl Hybrid
    #     (GCH) model and the control state of
    #     :py:meth:`~.GaussianModel.calculate_VW_velocities` and
    #     :py:attr:`~.VelocityDeflection.use_secondary_steering`.

    #     Args:
    #         enable (bool, optional): Flag whether or not to implement secondary
    #         steering from GCH model. Defaults to *True*.
    #     """
    #     model_params = self.get_model_parameters()
    #     use_yaw_added_recovery = model_params["Wake Velocity Parameters"]["use_yaw_added_recovery"]

    #     if enable:
    #         model_params["Wake Deflection Parameters"]["use_secondary_steering"] = True

    #         # If enabling be sure calc vw is on
    #         model_params["Wake Velocity Parameters"]["calculate_VW_velocities"] = True

    #     if not enable:
    #         model_params["Wake Deflection Parameters"]["use_secondary_steering"] = False

    #         # If yar is also off, disable calculate_VW_velocities
    #         if not use_yaw_added_recovery:
    #             model_params["Wake Velocity Parameters"]["calculate_VW_velocities"] = False

    #     self.set_model_parameters(model_params)
    #     self.reinitialize()

    # def show_model_parameters(
    #     self,
    #     params=None,
    #     verbose=False,
    #     wake_velocity_model=True,
    #     wake_deflection_model=True,
    #     turbulence_model=False,
    # ):
    #     """
    #     Helper function to print the current wake model parameters and values.
    #     Shortcut to :py:meth:`~.tools.interface_utilities.show_params`.

    #     Args:
    #         params (list, optional): Specific model parameters to be returned,
    #             supplied as a list of strings. If None, then returns all
    #             parameters. Defaults to None.
    #         verbose (bool, optional): If set to *True*, will return the
    #             docstrings for each parameter. Defaults to *False*.
    #         wake_velocity_model (bool, optional): If set to *True*, will return
    #             parameters from the wake_velocity model. If set to *False*, will
    #             exclude parameters from the wake velocity model. Defaults to
    #             *True*.
    #         wake_deflection_model (bool, optional): If set to *True*, will
    #             return parameters from the wake deflection model. If set to
    #             *False*, will exclude parameters from the wake deflection
    #             model. Defaults to *True*.
    #         turbulence_model (bool, optional): If set to *True*, will return
    #             parameters from the wake turbulence model. If set to *False*,
    #             will exclude parameters from the wake turbulence model.
    #             Defaults to *True*.
    #     """
    #     show_params(
    #         self.floris.wake,
    #         params,
    #         verbose,
    #         wake_velocity_model,
    #         wake_deflection_model,
    #         turbulence_model,
    #     )

    # def get_model_parameters(
    #     self,
    #     params=None,
    #     wake_velocity_model=True,
    #     wake_deflection_model=True,
    #     turbulence_model=True,
    # ):
    #     """
    #     Helper function to return the current wake model parameters and values.
    #     Shortcut to :py:meth:`~.tools.interface_utilities.get_params`.

    #     Args:
    #         params (list, optional): Specific model parameters to be returned,
    #             supplied as a list of strings. If None, then returns all
    #             parameters. Defaults to None.
    #         wake_velocity_model (bool, optional): If set to *True*, will return
    #             parameters from the wake_velocity model. If set to *False*, will
    #             exclude parameters from the wake velocity model. Defaults to
    #             *True*.
    #         wake_deflection_model (bool, optional): If set to *True*, will
    #             return parameters from the wake deflection model. If set to
    #             *False*, will exclude parameters from the wake deflection
    #             model. Defaults to *True*.
    #         turbulence_model ([type], optional): If set to *True*, will return
    #             parameters from the wake turbulence model. If set to *False*,
    #             will exclude parameters from the wake turbulence model.
    #             Defaults to *True*.

    #     Returns:
    #         dict: Dictionary containing model parameters and their values.
    #     """
    #     model_params = get_params(
    #         self.floris.wake, params, wake_velocity_model, wake_deflection_model, turbulence_model
    #     )

    #     return model_params

    # def set_model_parameters(self, params, verbose=True):
    #     """
    #     Helper function to set current wake model parameters.
    #     Shortcut to :py:meth:`~.tools.interface_utilities.set_params`.

    #     Args:
    #         params (dict): Specific model parameters to be set, supplied as a
    #             dictionary of key:value pairs.
    #         verbose (bool, optional): If set to *True*, will print information
    #             about each model parameter that is changed. Defaults to *True*.
    #     """
    #     self.floris.wake = set_params(self.floris.wake, params, verbose)






    # def vis_layout(
    #     self,
    #     ax=None,
    #     show_wake_lines=False,
    #     limit_dist=None,
    #     turbine_face_north=False,
    #     one_index_turbine=False,
    #     black_and_white=False,
    # ):
    #     """
    #     Visualize the layout of the wind farm in the floris instance.
    #     Shortcut to :py:meth:`~.tools.layout_functions.visualize_layout`.

    #     Args:
    #         ax (:py:class:`matplotlib.pyplot.axes`, optional):
    #             Figure axes. Defaults to None.
    #         show_wake_lines (bool, optional): Flag to control plotting of
    #             wake boundaries. Defaults to False.
    #         limit_dist (float, optional): Downstream limit to plot wakes.
    #             Defaults to None.
    #         turbine_face_north (bool, optional): Force orientation of wind
    #             turbines. Defaults to False.
    #         one_index_turbine (bool, optional): If *True*, 1st turbine is
    #             turbine 1.
    #     """
    #     for i, turbine in enumerate(self.floris.farm.turbines):
    #         D = turbine.rotor_diameter
    #         break
    #     layout_x, layout_y = self.get_turbine_layout()

    #     turbineLoc = build_turbine_loc(layout_x, layout_y)

    #     # Show visualize the turbine layout
    #     visualize_layout(
    #         turbineLoc,
    #         D,
    #         ax=ax,
    #         show_wake_lines=show_wake_lines,
    #         limit_dist=limit_dist,
    #         turbine_face_north=turbine_face_north,
    #         one_index_turbine=one_index_turbine,
    #         black_and_white=black_and_white,
    #     )

    # def show_flow_field(self, ax=None):
    #     """
    #     Shortcut method to
    #     :py:meth:`~.tools.visualization.visualize_cut_plane`.

    #     Args:
    #         ax (:py:class:`matplotlib.pyplot.axes` optional):
    #             Figure axes. Defaults to None.
    #     """
    #     # Get horizontal plane at default height (hub-height)
    #     hor_plane = self.get_hor_plane()

    #     # Plot and show
    #     if ax is None:
    #         fig, ax = plt.subplots()
    #     visualize_cut_plane(hor_plane, ax=ax)
    #     plt.show()



    ## Functionality removed in v3

    def set_rotor_diameter(self, rotor_diameter):
        """
        This function has been replaced and no longer works correctly, assigning an error
        """
        raise Exception(
            "FlorinInterface.set_rotor_diameter has been removed in favor of FlorinInterface.change_turbine. See examples/change_turbine/."
        )
